"""
Time series analysis functions.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import inspect, functools, weakref, glob, types

from yt.funcs import *
from yt.convenience import load
from .data_containers import data_object_registry
from .analyzer_objects import create_quantity_proxy, \
    analysis_task_registry, AnalysisTask
from .derived_quantities import quantity_info
from yt.utilities.exceptions import YTException
from yt.utilities.parallel_tools.parallel_analysis_interface \
    import parallel_objects, parallel_root_only
from yt.utilities.parameter_file_storage import \
    simulation_time_series_registry

class AnalysisTaskProxy(object):
    def __init__(self, time_series):
        self.time_series = time_series

    def __getitem__(self, key):
        task_cls = analysis_task_registry[key]
        @wraps(task_cls.__init__)
        def func(*args, **kwargs):
            task = task_cls(*args, **kwargs)
            return self.time_series.eval(task)
        return func

    def keys(self):
        return analysis_task_registry.keys()

    def __contains__(self, key):
        return key in analysis_task_registry

def get_pf_prop(propname):
    def _eval(params, pf):
        return getattr(pf, propname)
    cls = type(propname, (AnalysisTask,),
                dict(eval = _eval, _params = tuple()))
    return cls

attrs = ("refine_by", "dimensionality", "current_time",
         "domain_dimensions", "domain_left_edge",
         "domain_right_edge", "unique_identifier",
         "current_redshift", "cosmological_simulation",
         "omega_matter", "omega_lambda", "hubble_constant")

class TimeSeriesParametersContainer(object):
    def __init__(self, data_object):
        self.data_object = data_object

    def __getattr__(self, attr):
        if attr in attrs:
            return self.data_object.eval(get_pf_prop(attr)())
        raise AttributeError(attr)

class TimeSeriesData(object):
    r"""The TimeSeriesData object is a container of multiple datasets,
    allowing easy iteration and computation on them.

    TimeSeriesData objects are designed to provide easy ways to access,
    analyze, parallelize and visualize multiple datasets sequentially.  This is
    primarily expressed through iteration, but can also be constructed via
    analysis tasks (see :ref:`time-series-analysis`).

    The best method to construct TimeSeriesData objects is through 
    :meth:`~yt.data_objects.time_series.TimeSeriesData.from_filenames`.


    Examples
    --------

    >>> ts = TimeSeriesData.from_filenames(
            "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0[0-6][0-9]0")
    >>> for pf in ts:
    ...     SlicePlot(pf, "x", "Density").save()

    """
    def __init__(self, outputs, parallel = True ,**kwargs):
        self.tasks = AnalysisTaskProxy(self)
        self.params = TimeSeriesParametersContainer(self)
        self._pre_outputs = outputs[:]
        for type_name in data_object_registry:
            setattr(self, type_name, functools.partial(
                TimeSeriesDataObject, self, type_name))
        self.parallel = parallel
        self.kwargs = kwargs

    def __iter__(self):
        # We can make this fancier, but this works
        for o in self._pre_outputs:
            if isinstance(o, types.StringTypes):
                yield load(o,**self.kwargs)
            else:
                yield o

    def __getitem__(self, key):
        if isinstance(key, types.SliceType):
            if isinstance(key.start, types.FloatType):
                return self.get_range(key.start, key.stop)
            # This will return a sliced up object!
            return TimeSeriesData(self._pre_outputs[key], self.parallel)
        o = self._pre_outputs[key]
        if isinstance(o, types.StringTypes):
            o = load(o,**self.kwargs)
        return o

    def __len__(self):
        return len(self._pre_outputs)

    def piter(self, storage = None):
        r"""Iterate over time series components in parallel.

        This allows you to iterate over a time series while dispatching
        individual components of that time series to different processors or
        processor groups.  If the parallelism strategy was set to be
        multi-processor (by "parallel = N" where N is an integer when the
        TimeSeriesData was created) this will issue each dataset to an
        N-processor group.  For instance, this would allow you to start a 1024
        processor job, loading up 100 datasets in a time series and creating 8
        processor groups of 128 processors each, each of which would be
        assigned a different dataset.  This could be accomplished as shown in
        the examples below.  The *storage* option is as seen in
        :func:`~yt.utilities.parallel_tools.parallel_analysis_interface.parallel_objects`
        which is a mechanism for storing results of analysis on an individual
        dataset and then combining the results at the end, so that the entire
        set of processors have access to those results.

        Note that supplying a *store* changes the iteration mechanism; see
        below.

        Parameters
        ----------
        storage : dict
            This is a dictionary, which will be filled with results during the
            course of the iteration.  The keys will be the parameter file
            indices and the values will be whatever is assigned to the *result*
            attribute on the storage during iteration.

        Examples
        --------
        Here is an example of iteration when the results do not need to be
        stored.  One processor will be assigned to each parameter file.

        >>> ts = TimeSeriesData.from_filenames("DD*/DD*.hierarchy")
        >>> for pf in ts.piter():
        ...    SlicePlot(pf, "x", "Density").save()
        ...
        
        This demonstrates how one might store results:

        >>> ts = TimeSeriesData.from_filenames("DD*/DD*.hierarchy")
        >>> my_storage = {}
        >>> for sto, pf in ts.piter(storage=my_storage):
        ...     v, c = pf.h.find_max("Density")
        ...     sto.result = (v, c)
        ...
        >>> for i, (v, c) in sorted(my_storage.items()):
        ...     print "% 4i  %0.3e" % (i, v)
        ...

        This shows how to dispatch 4 processors to each dataset:

        >>> ts = TimeSeriesData.from_filenames("DD*/DD*.hierarchy",
        ...                     parallel = 4)
        >>> for pf in ts.piter():
        ...     ProjectionPlot(pf, "x", "Density").save()
        ...

        """
        dynamic = False
        if self.parallel == False:
            njobs = 1
        else:
            if self.parallel == True: njobs = -1
            else: njobs = self.parallel
        return parallel_objects(self, njobs=njobs, storage=storage,
                                dynamic=dynamic)

    def eval(self, tasks, obj=None):
        tasks = ensure_list(tasks)
        return_values = {}
        for store, pf in self.piter(return_values):
            store.result = []
            for task in tasks:
                try:
                    style = inspect.getargspec(task.eval)[0][1]
                    if style == 'pf':
                        arg = pf
                    elif style == 'data_object':
                        if obj == None:
                            obj = TimeSeriesDataObject(self, "all_data")
                        arg = obj.get(pf)
                    rv = task.eval(arg)
                # We catch and store YT-originating exceptions
                # This fixes the standard problem of having a sphere that's too
                # small.
                except YTException as rv:
                    pass
                store.result.append(rv)
        return [v for k, v in sorted(return_values.items())]

    @classmethod
    def from_filenames(cls, filenames, parallel = True, **kwargs):
        r"""Create a time series from either a filename pattern or a list of
        filenames.

        This method provides an easy way to create a
        :class:`~yt.data_objects.time_series.TimeSeriesData`, given a set of
        filenames or a pattern that matches them.  Additionally, it can set the
        parallelism strategy.

        Parameters
        ----------
        filenames : list or pattern
            This can either be a list of filenames (such as ["DD0001/DD0001",
            "DD0002/DD0002"]) or a pattern to match, such as
            "DD*/DD*.hierarchy").  If it's the former, they will be loaded in
            order.  The latter will be identified with the glob module and then
            sorted.
        parallel : True, False or int
            This parameter governs the behavior when .piter() is called on the
            resultant TimeSeriesData object.  If this is set to False, the time
            series will not iterate in parallel when .piter() is called.  If
            this is set to either True or an integer, it will be iterated with
            1 or that integer number of processors assigned to each parameter
            file provided to the loop.

        Examples
        --------

        >>> ts = TimeSeriesData.from_filenames(
                "GasSloshingLowRes/sloshing_low_res_hdf5_plt_cnt_0[0-6][0-9]0")
        >>> for pf in ts:
        ...     SlicePlot(pf, "x", "Density").save()

        """
        
        if isinstance(filenames, types.StringTypes):
            filenames = glob.glob(filenames)
            filenames.sort()
        if len(filenames) == 0:
            raise YTOutputNotIdentified(filenames, {})
        obj = cls(filenames[:], parallel = parallel, **kwargs)
        return obj

    @classmethod
    def from_output_log(cls, output_log,
                        line_prefix = "DATASET WRITTEN",
                        parallel = True):
        filenames = []
        for line in open(output_log):
            if not line.startswith(line_prefix): continue
            cut_line = line[len(line_prefix):].strip()
            fn = cut_line.split()[0]
            filenames.append(fn)
        obj = cls(filenames, parallel = parallel)
        return obj

class TimeSeriesQuantitiesContainer(object):
    def __init__(self, data_object, quantities):
        self.data_object = data_object
        self.quantities = quantities

    def __getitem__(self, key):
        if key not in self.quantities: raise KeyError(key)
        q = self.quantities[key]
        def run_quantity_wrapper(quantity, quantity_name):
            @wraps(quantity_info[quantity_name][1])
            def run_quantity(*args, **kwargs):
                to_run = quantity(*args, **kwargs)
                return self.data_object.eval(to_run)
            return run_quantity
        return run_quantity_wrapper(q, key)

class TimeSeriesDataObject(object):
    def __init__(self, time_series, data_object_name, *args, **kwargs):
        self.time_series = weakref.proxy(time_series)
        self.data_object_name = data_object_name
        self._args = args
        self._kwargs = kwargs
        qs = dict([(qn, create_quantity_proxy(qv)) for qn, qv in quantity_info.items()])
        self.quantities = TimeSeriesQuantitiesContainer(self, qs)

    def eval(self, tasks):
        return self.time_series.eval(tasks, self)

    def get(self, pf):
        # We get the type name, which corresponds to an attribute of the
        # hierarchy
        cls = getattr(pf.h, self.data_object_name)
        return cls(*self._args, **self._kwargs)


class SimulationTimeSeries(TimeSeriesData):
    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            code_name = name[:name.find('Simulation')]
            if code_name:
                simulation_time_series_registry[code_name] = cls
                mylog.debug("Registering simulation: %s as %s", code_name, cls)

    def __init__(self, parameter_filename, find_outputs=False):
        """
        Base class for generating simulation time series types.
        Principally consists of a *parameter_filename*.
        """

        if not os.path.exists(parameter_filename):
            raise IOError(parameter_filename)
        self.parameter_filename = parameter_filename
        self.basename = os.path.basename(parameter_filename)
        self.directory = os.path.dirname(parameter_filename)
        self.parameters = {}

        # Set some parameter defaults.
        self._set_parameter_defaults()
        # Read the simulation parameter file.
        self._parse_parameter_file()
        # Set up time units dictionary.
        self._set_time_units()

        # Figure out the starting and stopping times and redshift.
        self._calculate_simulation_bounds()
        # Get all possible datasets.
        self._get_all_outputs(find_outputs=find_outputs)
        
        self.print_key_parameters()

    def __repr__(self):
        return self.parameter_filename

    @parallel_root_only
    def print_key_parameters(self):
        """
        Print out some key parameters for the simulation.
        """
        for a in ["domain_dimensions", "domain_left_edge",
                  "domain_right_edge", "initial_time", "final_time",
                  "stop_cycle", "cosmological_simulation"]:
            if not hasattr(self, a):
                mylog.error("Missing %s in parameter file definition!", a)
                continue
            v = getattr(self, a)
            mylog.info("Parameters: %-25s = %s", a, v)
        if hasattr(self, "cosmological_simulation") and \
           getattr(self, "cosmological_simulation"):
            for a in ["box_size", "omega_lambda",
                      "omega_matter", "hubble_constant",
                      "initial_redshift", "final_redshift"]:
                if not hasattr(self, a):
                    mylog.error("Missing %s in parameter file definition!", a)
                    continue
                v = getattr(self, a)
                mylog.info("Parameters: %-25s = %s", a, v)
        mylog.info("Total datasets: %d." % len(self.all_outputs))




"""
This is a library of yt-defined exceptions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# We don't need to import 'exceptions'
#import exceptions
import os.path

class YTException(Exception):
    def __init__(self, pf = None):
        Exception.__init__(self)
        self.pf = pf

# Data access exceptions:

class YTOutputNotIdentified(YTException):
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return "Supplied %s %s, but could not load!" % (
            self.args, self.kwargs)

class YTSphereTooSmall(YTException):
    def __init__(self, pf, radius, smallest_cell):
        YTException.__init__(self, pf)
        self.radius = radius
        self.smallest_cell = smallest_cell

    def __str__(self):
        return "%0.5e < %0.5e" % (self.radius, self.smallest_cell)

class YTAxesNotOrthogonalError(YTException):
    def __init__(self, axes):
        self.axes = axes

    def __str__(self):
        return "The supplied axes are not orthogonal.  %s" % (self.axes)

class YTNoDataInObjectError(YTException):
    def __init__(self, obj):
        self.obj_type = getattr(obj, "_type_name", "")

    def __str__(self):
        s = "The object requested has no data included in it."
        if self.obj_type == "slice":
            s += "  It may lie on a grid face.  Try offsetting slightly."
        return s

class YTSimulationNotIdentified(YTException):
    def __init__(self, sim_type):
        YTException.__init__(self)
        self.sim_type = sim_type

    def __str__(self):
        return "Simulation time-series type %s not defined." % self.sim_type

class YTCannotParseFieldDisplayName(YTException):
    def __init__(self, field_name, display_name, mathtext_error):
        self.field_name = field_name
        self.display_name = display_name
        self.mathtext_error = mathtext_error

    def __str__(self):
        return ("The display name \"%s\" "
                "of the derived field %s " 
                "contains the following LaTeX parser errors:\n" ) \
                % (self.display_name, self.field_name) + self.mathtext_error

class YTCannotParseUnitDisplayName(YTException):
    def __init__(self, field_name, display_unit, mathtext_error):
        self.field_name = field_name
        self.unit_name = unit_name
        self.mathtext_error = mathtext_error

    def __str__(self):
        return ("The unit display name \"%s\" "
                "of the derived field %s " 
                "contains the following LaTeX parser errors:\n" ) \
            % (self.unit_name, self.field_name) + self.mathtext_error

class InvalidSimulationTimeSeries(YTException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message
            
class MissingParameter(YTException):
    def __init__(self, pf, parameter):
        YTException.__init__(self, pf)
        self.parameter = parameter

    def __str__(self):
        return "Parameter file %s is missing %s parameter." % \
            (self.pf, self.parameter)

class NoStoppingCondition(YTException):
    def __init__(self, pf):
        YTException.__init__(self, pf)

    def __str__(self):
        return "Simulation %s has no stopping condition.  StopTime or StopCycle should be set." % \
            self.pf

class YTNotInsideNotebook(YTException):
    def __str__(self):
        return "This function only works from within an IPython Notebook."

class YTNotDeclaredInsideNotebook(YTException):
    def __str__(self):
        return "You have not declared yourself to be inside the IPython" + \
               "Notebook.  Do so with this command:\n\n" + \
               "ytcfg['yt','ipython_notebook'] = 'True'"

class YTUnitNotRecognized(YTException):
    def __init__(self, unit):
        self.unit = unit

    def __str__(self):
        return "This parameter file doesn't recognize %s" % self.unit

class YTHubRegisterError(YTException):
    def __str__(self):
        return "You must create an API key before uploading.  See " + \
               "https://data.yt-project.org/getting_started.html"

class YTNoFilenamesMatchPattern(YTException):
    def __init__(self, pattern):
        self.pattern = pattern

    def __str__(self):
        return "No filenames were found to match the pattern: " + \
               "'%s'" % (self.pattern)

class YTNoOldAnswer(YTException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "There is no old answer available.\n" + \
               str(self.path)

class YTCloudError(YTException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "Failed to retrieve cloud data. Connection may be broken.\n" + \
               str(self.path)

class YTEllipsoidOrdering(YTException):
    def __init__(self, pf, A, B, C):
        YTException.__init__(self, pf)
        self._A = A
        self._B = B
        self._C = C

    def __str__(self):
        return "Must have A>=B>=C"

class EnzoTestOutputFileNonExistent(YTException):
    def __init__(self, filename):
        self.filename = filename
        self.testname = os.path.basename(os.path.dirname(filename))

    def __str__(self):
        return "Enzo test output file (OutputLog) not generated for: " + \
            "'%s'" % (self.testname) + ".\nTest did not complete."

class YTNoAPIKey(YTException):
    def __init__(self, service, config_name):
        self.service = service
        self.config_name = config_name

    def __str__(self):
        return "You need to set an API key for %s in ~/.yt/config as %s" % (
            self.service, self.config_name)

class YTTooManyVertices(YTException):
    def __init__(self, nv, fn):
        self.nv = nv
        self.fn = fn

    def __str__(self):
        s = "There are too many vertices (%s) to upload to Sketchfab. " % (self.nv)
        s += "Your model has been saved as %s .  You should upload manually." % (self.fn)
        return s

class YTInvalidWidthError(YTException):
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return str(self.error)



"""
Enzo-specific IO functions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict

import exceptions
import os

from yt.utilities.io_handler import \
    BaseIOHandler, _axis_ids
from yt.utilities.logger import ytLogger as mylog

class IOHandlerStream(BaseIOHandler):

    _data_style = "stream"

    def __init__(self, stream_handler):
        self.fields = stream_handler.fields
        BaseIOHandler.__init__(self)

    def _read_data(self, grid, field):
        # This is where we implement processor-locking
        #if grid.id not in self.grids_in_memory:
        #    mylog.error("Was asked for %s but I have %s", grid.id, self.grids_in_memory.keys())
        #    raise KeyError
        tr = self.fields[grid.id][field].copy()
        # If it's particles, we copy.
        return tr

    def modify(self, field):
        return field

    def _read_field_names(self, grid):
        return self.fields[grid.id].keys()

    @property
    def _read_exception(self):
        return KeyError




"""
Shareable definitions for common fp/int Cython utilities

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

cimport numpy as np
cimport cython

cdef inline int imax(int i0, int i1) nogil:
    if i0 > i1: return i0
    return i1

cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) nogil:
    if f0 > f1: return f0
    return f1

cdef inline int imin(int i0, int i1) nogil:
    if i0 < i1: return i0
    return i1

cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) nogil:
    if f0 < f1: return f0
    return f1

cdef inline np.float64_t fabs(np.float64_t f0) nogil:
    if f0 < 0.0: return -f0
    return f0

cdef inline int iclip(int i, int a, int b) nogil:
    if i < a: return a
    if i > b: return b
    return i

cdef inline int i64clip(np.int64_t i, np.int64_t a, np.int64_t b) nogil:
    if i < a: return a
    if i > b: return b
    return i

cdef inline np.float64_t fclip(np.float64_t f,
                      np.float64_t a, np.float64_t b) nogil:
    return fmin(fmax(f, a), b)




"""
A simple leaflet-based pannable map server

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import numpy as np
import zipfile
import sys

from yt.visualization.image_writer import apply_colormap
from yt.visualization.fixed_resolution import FixedResolutionBuffer
from yt.utilities.lib import write_png_to_string, get_color_bounds

import yt.utilities.bottle as bottle

from yt.funcs import *
local_dir = os.path.dirname(__file__)

def exc_writeout(f):
    import traceback
    @wraps(f)
    def func(*args, **kwargs):
        try:
            rv = f(*args, **kwargs)
            return rv
        except Exception as e:
            traceback.print_exc(None, open("temp.exc", "w"))
            raise
    return func

class PannableMapServer(object):
    _widget_name = "pannable_map"
    reasonjs_file = None
    def __init__(self, data, field, route_prefix = ""):
        self.data = data
        self.pf = data.pf
        self.field = field
        
        bottle.route("%s/map/:L/:x/:y.png" % route_prefix)(self.map)
        bottle.route("%s/" % route_prefix)(self.index)
        bottle.route("%s/index.html" % route_prefix)(self.index)
        # This is a double-check, since we do not always mandate this for
        # slices:
        self.data[self.field] = self.data[self.field].astype("float64")
        if route_prefix == "":
            # We assume this means we're running standalone
            from .utils import get_reasonjs_path
            try:
                reasonjs_path = get_reasonjs_path()
            except IOError:
                sys.exit(1)
            self.reasonjs_file = zipfile.ZipFile(reasonjs_path, 'r')
            bottle.route("/reason-js/:path#.+#", "GET")(self.static)

    def map(self, L, x, y):
        dd = 1.0 / (2.0**(int(L)))
        relx = int(x) * dd
        rely = int(y) * dd
        DW = (self.pf.domain_right_edge - self.pf.domain_left_edge)
        xl = self.pf.domain_left_edge[0] + relx * DW[0]
        yl = self.pf.domain_left_edge[1] + rely * DW[1]
        xr = xl + dd*DW[0]
        yr = yl + dd*DW[1]
        frb = FixedResolutionBuffer(self.data, (xl, xr, yl, yr), (256, 256))
        cmi, cma = get_color_bounds(self.data['px'], self.data['py'],
                                    self.data['pdx'], self.data['pdy'],
                                    self.data[self.field],
                                    self.pf.domain_left_edge[0],
                                    self.pf.domain_right_edge[0],
                                    self.pf.domain_left_edge[1],
                                    self.pf.domain_right_edge[1],
                                    dd*DW[0] / (64*256),
                                    dd*DW[0])
        if self.pf.field_info[self.field].take_log:
            cmi = np.log10(cmi)
            cma = np.log10(cma)
            to_plot = apply_colormap(np.log10(frb[self.field]), color_bounds = (cmi, cma))
        else:
            to_plot = apply_colormap(frb[self.field], color_bounds = (cmi, cma))
        rv = write_png_to_string(to_plot)
        return rv

    def index(self):
        return bottle.static_file("map_index.html",
                    root=os.path.join(local_dir, "html"))

    def static(self, path):
        if self.reasonjs_file is None: raise RuntimeError
        pp = os.path.join("reason-js", path)
        try:
            f = self.reasonjs_file.open(pp)
        except KeyError:
            bottle.response.status = 404
            return
        if path[-4:].lower() in (".png", ".gif", ".jpg"):
            bottle.response.headers['Content-Type'] = "image/%s" % (path[-3:].lower())
        elif path[-4:].lower() == ".css":
            bottle.response.headers['Content-Type'] = "text/css"
        elif path[-3:].lower() == ".js":
            bottle.response.headers['Content-Type'] = "text/javascript"
        return f.read()



"""
The data-file handling functions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict

import yt.utilities.lib as au
import exceptions
import cPickle
import os
import h5py

_axis_ids = {0:2,1:1,2:0}

io_registry = {}

class BaseIOHandler(object):

    _data_style = None
    _particle_reader = False

    class __metaclass__(type):
        def __init__(cls, name, b, d):
            type.__init__(cls, name, b, d)
            if hasattr(cls, "_data_style"):
                io_registry[cls._data_style] = cls

    def __init__(self):
        self.queue = defaultdict(dict)

    # We need a function for reading a list of sets
    # and a function for *popping* from a queue all the appropriate sets

    def preload(self, grids, sets):
        pass

    def pop(self, grid, field):
        if grid.id in self.queue and field in self.queue[grid.id]:
            return self.modify(self.queue[grid.id].pop(field))
        else:
            # We only read the one set and do not store it if it isn't pre-loaded
            return self._read_data_set(grid, field)

    def peek(self, grid, field):
        return self.queue[grid.id].get(field, None)

    def push(self, grid, field, data):
        if grid.id in self.queue and field in self.queue[grid.id]:
            raise ValueError
        self.queue[grid][field] = data

    def _field_in_backup(self, grid, backup_file, field_name):
        if os.path.exists(backup_file):
            fhandle = h5py.File(backup_file, 'r')
            g = fhandle["data"]
            grid_group = g["grid_%010i" % (grid.id - grid._id_offset)]
            if field_name in grid_group:
                return_val = True
            else:
                return_val = False
            fhandle.close()
            return return_val
        else:
            return False
            
    def _read_data_set(self, grid, field):
        # check backup file first. if field not found,
        # call frontend-specific io method
        backup_filename = grid.pf.backup_filename
        if not grid.pf.read_from_backup:
            return self._read_data(grid, field)
        elif self._field_in_backup(grid, backup_filename, field):
            fhandle = h5py.File(backup_filename, 'r')
            g = fhandle["data"]
            grid_group = g["grid_%010i" % (grid.id - grid._id_offset)]
            data = grid_group[field][:]
            fhandle.close()
            return data
        else:
            return self._read_data(grid, field)
                
    # Now we define our interface
    def _read_data(self, grid, field):
        pass

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        tr = self._read_data_set(grid, field)[sl]
        if tr.dtype == "float32": tr = tr.astype("float64")
        return tr

    def _read_field_names(self, grid):
        pass

    @property
    def _read_exception(self):
        return None

class IOHandlerExtracted(BaseIOHandler):

    _data_style = 'extracted'

    def _read_data_set(self, grid, field):
        return (grid.base_grid[field] / grid.base_grid.convert(field))

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        return grid.base_grid[field][tuple(sl)] / grid.base_grid.convert(field)



"""
Task queue for reason.

Author: Britton Smith <brittonsmith@gmail.com>
Affiliation: Michigan State University
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from yt.funcs import *

from yt.gui.reason.basic_repl import ProgrammaticREPL
from yt.gui.reason.extdirect_repl import ExecutionThread
from yt.gui.reason.bottle_mods import PayloadHandler
from .utils import get_list_of_datasets

class PyroQueueRoot(object):
    def __init__(self, comm):
        self.comm = comm
        self.repl = ProgrammaticREPL()
        self.execution_thread = ExecutionThread(self.repl)
        self.payload_handler = PayloadHandler()
        self.execution_thread.start()

    def execute(self, code, hide = False):
        mylog.info('Root sending out code.')
        code = self.comm.comm.bcast(code, root=0)
        task = {'type': 'code',
                'code': code,
                'hide': hide}
        self.execution_thread.queue.put(task)

    def deliver(self):
        return self.payload_handler.deliver_payloads()

class PyroQueueNonRoot(object):
    def __init__(self, comm):
        self.comm = comm
        self.repl = ProgrammaticREPL()

    def run(self):
        while 1:
            code = None
            code = self.comm.comm.bcast(code, root=0)
            mylog.info('Received commands from subcomm root.')
            value = self.repl.execute(code)
            



#!/usr/bin/env python
import setuptools
import os
import sys
import os.path

def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('halo_finding', parent_package, top_path)
    config.add_subpackage("fof")
    config.add_subpackage("hop")
    config.add_subpackage("parallel_hop")
    if os.path.exists("rockstar.cfg"):
        config.add_subpackage("rockstar")
    config.make_config_py() # installs __config__.py
    #config.make_svn_version_py()
    return config



"""Delaunay triangulation and interpolation tools.

:Author: Robert Kern <robert.kern@gmail.com>
:Copyright: Copyright 2005 Robert Kern.
:License: BSD-style license. See LICENSE.txt
"""

from _delaunay import delaunay
from triangulate import *
from interpolate import *



"""
Turn on and off perftools profiling

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# For more info:
# https://pygabriel.wordpress.com/2010/04/14/profiling-python-c-extensions/

# prof.pyx
cdef extern from "google/profiler.h":
    void ProfilerStart( char* fname )
    void ProfilerStop()

def profiler_start(fname):
    ProfilerStart(<char *>fname)

def profiler_stop():
    ProfilerStop()




"""
Function to calculate volume in common between two n-cubes, with optional
periodic boundary conditions.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

def common_volume(n_cube_1, n_cube_2, periodic=None):
    "Return the n-volume in common between the two n-cubes."

    # Check for proper args.
    if ((len(np.shape(n_cube_1)) != 2) or
        (np.shape(n_cube_1)[1] != 2) or
        (np.shape(n_cube_1) != np.shape(n_cube_2))):
        print "Arguments must be 2 (n, 2) numpy array."
        return 0

    if ((periodic is not None) and
        (np.shape(n_cube_1) != np.shape(periodic))):
        print "periodic argument must be (n, 2) numpy array."
        return 0

    nCommon = 1.0
    for q in range(np.shape(n_cube_1)[0]):
        if (periodic is None):
            nCommon *= common_segment(n_cube_1[q], n_cube_2[q])
        else:
            nCommon *= common_segment(n_cube_1[q], n_cube_2[q],
                                      periodic=periodic[q])

    return nCommon

def common_segment(seg1, seg2, periodic=None):
    "Return the length of the common segment."

    # Check for proper args.
    if ((len(seg1) != 2) or (len(seg2) != 2)):
        print "Arguments must be arrays of size 2."
        return 0

    # If not periodic, then this is very easy.
    if periodic is None:
        seg1.sort()
        len1 = seg1[1] - seg1[0]
        seg2.sort()
        len2 = seg2[1] - seg2[0]

        common = 0.0

        add = seg1[1] - seg2[0]
        if ((add > 0) and (add <= max(len1, len2))):
            common += add
        add = seg2[1] - seg1[0]
        if ((add > 0) and (add <= max(len1, len2))):
            common += add
        common = min(common, len1, len2)
        return common

    # If periodic, it's a little more complicated.
    else:
        if len(periodic) != 2:
            print "periodic array must be of size 2."
            return 0

        seg1.sort()
        flen1 = seg1[1] - seg1[0]
        len1 = flen1 - int(flen1)
        seg2.sort()
        flen2 = seg2[1] - seg2[0]
        len2 = flen2 - int(flen2)

        periodic.sort()
        scale = periodic[1] - periodic[0]

        if (abs(int(flen1)-int(flen2)) >= scale):
            return min(flen1, flen2)

        # Adjust for periodicity
        seg1[0] = np.mod(seg1[0], scale) + periodic[0]
        seg1[1] = seg1[0] + len1
        if (seg1[1] > periodic[1]): seg1[1] -= scale
        seg2[0] = np.mod(seg2[0], scale) + periodic[0]
        seg2[1] = seg2[0] + len2
        if (seg2[1] > periodic[1]): seg2[1] -= scale

        # create list of non-periodic segments
        pseg1 = []
        if (seg1[0] >= seg1[1]):
            pseg1.append([seg1[0], periodic[1]])
            pseg1.append([periodic[0], seg1[1]])
        else:
            pseg1.append(seg1)
        pseg2 = []
        if (seg2[0] >= seg2[1]):
            pseg2.append([seg2[0], periodic[1]])
            pseg2.append([periodic[0], seg2[1]])
        else:
            pseg2.append(seg2)

        # Add up common segments.
        common = min(int(flen1), int(flen2))

        for subseg1 in pseg1:
            for subseg2 in pseg2:
                common += common_segment(subseg1, subseg2)

        return common



"""
A read-eval-print-loop that is served up through Bottle and accepts its
commands through ExtDirect calls

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import json
import os
import stat
import cStringIO
import logging
import uuid
import numpy as np
import time
import urllib
import urllib2
import pprint
import traceback
import tempfile
import base64
import imp
import threading
import Queue
import zipfile
try:
    import Pyro4
except ImportError:
    pass

from yt.funcs import *
from yt.utilities.logger import ytLogger, ufstring
from yt.utilities.definitions import inv_axis_names
from yt.visualization.image_writer import apply_colormap
from yt.visualization.api import Streamlines
from .widget_store import WidgetStore

from .bottle_mods import preroute, BottleDirectRouter, notify_route, \
                         PayloadHandler, lockit
from yt.utilities.bottle import response, request, route, static_file
from .utils import get_list_of_datasets
from .basic_repl import ProgrammaticREPL

try:
    import pygments
    import pygments.lexers
    import pygments.formatters
    def _highlighter():
        pl = pygments.lexers.PythonLexer()
        hf = pygments.formatters.HtmlFormatter(
                linenos='table', linenospecial=2)
        def __highlighter(a):
            return pygments.highlight(a, pl, hf)
        return __highlighter, hf.get_style_defs()
    # We could add an additional '.highlight_pyg' in the call
    highlighter, highlighter_css = _highlighter()
except ImportError:
    highlighter = lambda a: a
    highlight_css = ''

local_dir = os.path.dirname(__file__)

class ExecutionThread(threading.Thread):
    def __init__(self, repl):
        self.repl = repl
        self.payload_handler = PayloadHandler()
        self.queue = Queue.Queue()
        threading.Thread.__init__(self)
        self.daemon = True

    def heartbeat(self):
        return

    def run(self):
        while 1:
            #print "Checking for a queue ..."
            try:
                task = self.queue.get(True, 1)
            except Queue.Empty:
                if self.repl.stopped: return
                continue
            print "Received the task", task
            if task['type'] != 'code':
                raise NotImplementedError
            print task
            self.execute_one(task['code'], task['hide'], task['result_id'])
            self.queue.task_done()

    def wait(self):
        self.queue.join()

    def execute_one(self, code, hide, result_id):
        self.repl.executed_cell_texts.append(code)
        result = ProgrammaticREPL.execute(self.repl, code)
        if self.repl.debug:
            print "==================== Cell Execution ===================="
            print code
            print "====================                ===================="
            print result
            print "========================================================"
        self.payload_handler.add_payload(
            {'type': 'cell',
             'output': result,
             'input': highlighter(code),
             'image_data': '',
             'result_id': result_id,
             'hide': hide,
             'raw_input': code},
            )
        objs = get_list_of_datasets()
        self.payload_handler.add_payload(
            {'type': 'dataobjects',
             'objs': objs})

class PyroExecutionThread(ExecutionThread):
    def __init__(self, repl):
        ExecutionThread.__init__(self, repl)
        hmac_key = raw_input("HMAC_KEY? ").strip()
        uri = raw_input("URI? ").strip()
        Pyro4.config.HMAC_KEY = hmac_key
        self.executor = Pyro4.Proxy(uri)

    def execute_one(self, code, hide, result_id):
        self.repl.executed_cell_texts.append(code)
        print code
        result = self.executor.execute(code)
        if not hide:
            self.repl.payload_handler.add_payload(
                {'type': 'cell',
                 'output': result,
                 'input': highlighter(code),
                 'result_id': result_id,
                 'raw_input': code},
                )
        ph = self.executor.deliver()
        for p in ph:
            self.repl.payload_handler.add_payload(p)

    def heartbeat(self):
        ph = self.executor.deliver()
        for p in ph:
            self.repl.payload_handler.add_payload(p)

def reason_pylab():
    from .utils import deliver_image
    def _canvas_deliver(canvas):
        tf = tempfile.TemporaryFile()
        canvas.print_png(tf)
        tf.seek(0)
        img_data = base64.b64encode(tf.read())
        tf.close()
        deliver_image(img_data)
    def reason_draw_if_interactive():
        if matplotlib.is_interactive():
            figManager =  Gcf.get_active()
            if figManager is not None:
                _canvas_deliver(figManager.canvas)
    def reason_show(mainloop = True):
        # We ignore mainloop here
        for manager in Gcf.get_all_fig_managers():
            _canvas_deliver(manager.canvas)
    # Matplotlib has very nice backend overriding.
    # We should really use that.  This is just a hack.
    import matplotlib
    new_agg = imp.new_module("reason_agg")
    import matplotlib.backends.backend_agg as bagg
    new_agg.__dict__.update(bagg.__dict__)
    new_agg.__dict__.update(
        {'show': reason_show,
         'draw_if_interactive': reason_draw_if_interactive})
    sys.modules["reason_agg"] = new_agg
    bagg.draw_if_interactive = reason_draw_if_interactive
    from matplotlib._pylab_helpers import Gcf
    import pylab, matplotlib
    matplotlib.rcParams["backend"] = "module://reason_agg"
    pylab.switch_backend("module://reason_agg")

_startup_template = r"""\
import pylab
from yt.mods import *
from yt.gui.reason.utils import load_script, deliver_image
from yt.gui.reason.widget_store import WidgetStore
from yt.data_objects.static_output import _cached_pfs

pylab.ion()
data_objects = []
widget_store = WidgetStore()
"""

class ExtDirectREPL(ProgrammaticREPL, BottleDirectRouter):
    _skip_expose = ('index')
    my_name = "ExtDirectREPL"
    timeout = 660 # a minute longer than the rocket server timeout
    server = None
    _heartbeat_timer = None

    def __init__(self, reasonjs_path, locals=None,
                 use_pyro=False):
        # First we do the standard initialization
        self.reasonjs_file = zipfile.ZipFile(reasonjs_path, 'r')
        ProgrammaticREPL.__init__(self, locals)
        # Now, since we want to only preroute functions we know about, and
        # since they have different arguments, and most of all because we only
        # want to add them to the routing tables (which are a singleton for the
        # entire interpreter state) we apply all the pre-routing now, rather
        # than through metaclasses or other fancy decorating.
        preroute_table = dict(index = ("/", "GET"),
                              _help_html = ("/help.html", "GET"),
                              _myapi = ("/ext-repl-api.js", "GET"),
                              _session_py = ("/session.py", "GET"),
                              _highlighter_css = ("/highlighter.css", "GET"),
                              _reasonjs = ("/reason-js/:path#.+#", "GET"),
                              _app = ("/reason/:path#.+#", "GET"),
                              )
        for v, args in preroute_table.items():
            preroute(args[0], method=args[1])(getattr(self, v))
        # This has to be routed to the root directory
        self.api_url = "repl"
        BottleDirectRouter.__init__(self)
        self.payload_handler = PayloadHandler()
        if use_pyro:
            self.execution_thread = PyroExecutionThread(self)
        else:
            self.execution_thread = ExecutionThread(self)
        # We pass in a reference to ourself
        self.execute(_startup_template)
        self.widget_store = WidgetStore(self)
        # Now we load up all the yt.mods stuff, but only after we've finished
        # setting up.
        reason_pylab()

    def activate(self):
        self.payload_handler._prefix = self._global_token
        self._setup_logging_handlers()
        # Setup our heartbeat
        self.last_heartbeat = time.time()
        self._check_heartbeat()
        self.execute("widget_store._global_token = '%s'" % self._global_token)
        self.execution_thread.start()

    def exception_handler(self, exc):
        result = {'type': 'cell',
                  'input': 'ERROR HANDLING IN REASON',
                  'result_id': None,
                  'output': traceback.format_exc()}
        return result

    def _setup_logging_handlers(self):
        handler = PayloadLoggingHandler()
        formatter = logging.Formatter(ufstring)
        handler.setFormatter(formatter)
        ytLogger.addHandler(handler)

    def index(self):
        root = os.path.join(local_dir, "html")
        return static_file("index.html", root)

    def heartbeat(self):
        self.last_heartbeat = time.time()
        if self.debug: print "### Heartbeat ... started: %s" % (time.ctime())
        for i in range(30):
            # Check for stop
            if self.debug: print "    ###"
            if self.stopped: return {'type':'shutdown'} # No race condition
            if self.payload_handler.event.wait(1): # One second timeout
                if self.debug: print "    ### Delivering payloads"
                rv = self.payload_handler.deliver_payloads()
                if self.debug: print "    ### Got back, returning"
                return rv
            self.execution_thread.heartbeat()
        if self.debug: print "### Heartbeat ... finished: %s" % (time.ctime())
        return []

    def _check_heartbeat(self):
        if self.server is not None:
            if not all((s._monitor.is_alive() for s in self.server.values())):
                self.shutdown()
                return
        if time.time() - self.last_heartbeat > self.timeout:
            print "Shutting down after a timeout of %s" % (self.timeout)
            #sys.exit(0)
            # Still can't shut down yet, because bottle doesn't return the
            # server instance by default.
            self.shutdown()
            return
        if self._heartbeat_timer is not None: return
        self._heartbeat_timer = threading.Timer(10, self._check_heartbeat)
        self._heartbeat_timer.start()

    def shutdown(self):
        if self.server is None:
            return
        self._heartbeat_timer.cancel()
        self.stopped = True
        self.payload_handler.event.set()
        for v in self.server.values():
            v.stop()
        for t in threading.enumerate():
            print "Found a living thread:", t

    def _help_html(self):
        root = os.path.join(local_dir, "html")
        return static_file("help.html", root)

    def _reasonjs(self, path):
        pp = os.path.join("reason-js", path)
        try:
            f = self.reasonjs_file.open(pp)
        except KeyError:
            response.status = 404
            return
        if path[-4:].lower() in (".png", ".gif", ".jpg"):
            response.headers['Content-Type'] = "image/%s" % (path[-3:].lower())
        elif path[-4:].lower() == ".css":
            response.headers['Content-Type'] = "text/css"
        elif path[-3:].lower() == ".js":
            response.headers['Content-Type'] = "text/javascript"
        return f.read()

    def _app(self, path):
        root = os.path.join(local_dir, "html")
        return static_file(path, root)

    def _highlighter_css(self):
        response.headers['Content-Type'] = "text/css"
        return highlighter_css

    def execute(self, code, hide = False, result_id = None):
        task = {'type': 'code',
                'code': code,
                'hide': hide,
                'result_id': result_id,
                }
        self.execution_thread.queue.put(task)
        return dict(status = True)

    def get_history(self):
        return self.executed_cell_texts[:]

    @lockit
    def save_session(self, filename):
        if filename.startswith('~'):
            filename = os.path.expanduser(filename)
        elif not filename.startswith('/'):
            filename = os.path.join(os.getcwd(), filename)
        if os.path.exists(filename):
            return {'status': 'FAIL', 'filename': filename,
                    'error': 'File exists!'}
        try:
            f = open(filename, 'w')
            f.write("\n######\n".join(self.executed_cell_texts))
            f.close()
        except IOError as (errno, strerror):
            return {'status': 'FAIL', 'filename': filename,
                    'error': strerror}
        except:
            return {'status': 'FAIL', 'filename': filename,
                    'error': 'Unexpected error.'}
        return {'status': 'SUCCESS', 'filename': filename}

    @lockit
    def paste_session(self):
        import xmlrpclib, cStringIO
        p = xmlrpclib.ServerProxy(
            "http://paste.yt-project.org/xmlrpc/",
            allow_none=True)
        cs = cStringIO.StringIO()
        cs.write("\n######\n".join(self.executed_cell_texts))
        cs = cs.getvalue()
        ret = p.pastes.newPaste('python', cs, None, '', '', True)
        site = "http://paste.yt-project.org/show/%s" % ret
        return {'status': 'SUCCESS', 'site': site}

    @lockit
    def paste_text(self, to_paste):
        import xmlrpclib, cStringIO
        p = xmlrpclib.ServerProxy(
            "http://paste.yt-project.org/xmlrpc/",
            allow_none=True)
        ret = p.pastes.newPaste('python', to_paste, None, '', '', True)
        site = "http://paste.yt-project.org/show/%s" % ret
        return {'status': 'SUCCESS', 'site': site}

    _api_key = 'f62d550859558f28c4c214136bc797c7'
    def upload_image(self, image_data, caption):
        if not image_data.startswith("data:"): return {'uploaded':False}
        prefix = "data:image/png;base64,"
        image_data = image_data[len(prefix):]
        parameters = {'key':self._api_key, 'image':image_data, type:'base64',
                      'caption': caption, 'title': "Uploaded Image from reason"}
        data = urllib.urlencode(parameters)
        req = urllib2.Request('http://api.imgur.com/2/upload.json', data)
        try:
            response = urllib2.urlopen(req).read()
        except urllib2.HTTPError as e:
            print "ERROR", e
            return {'uploaded':False}
        rv = json.loads(response)
        rv['uploaded'] = True
        pprint.pprint(rv)
        return rv

    @lockit
    def _session_py(self):
        cs = cStringIO.StringIO()
        cs.write("\n######\n".join(self.executed_cell_texts))
        cs.seek(0)
        response.headers["content-disposition"] = "attachment;"
        return cs

    @lockit
    def load(self, base_dir, filename):
        pp = os.path.join(base_dir, filename)
        funccall = "pfs.append(load('%s'))" % pp
        self.execute(funccall)
        return []

    def file_listing(self, base_dir, sub_dir):
        if base_dir == "":
            cur_dir = os.getcwd()
        elif sub_dir == "":
            cur_dir = base_dir
        else:
            cur_dir = os.path.join(base_dir, sub_dir)
            cur_dir = os.path.abspath(cur_dir)
        if not os.path.isdir(cur_dir):
            return {'change':False}
        fns = os.listdir(cur_dir)
        results = [("..", 0, "directory")]
        for fn in sorted((os.path.join(cur_dir, f) for f in fns)):
            if not os.access(fn, os.R_OK): continue
            if os.path.isfile(fn):
                size = os.path.getsize(fn)
                t = "file"
            else:
                size = 0
                t = "directory"
            results.append((os.path.basename(fn), size, t))
        return dict(objs = results, cur_dir=cur_dir)

class PayloadLoggingHandler(logging.StreamHandler):
    def __init__(self, *args, **kwargs):
        logging.StreamHandler.__init__(self, *args, **kwargs)
        self.payload_handler = PayloadHandler()

    def emit(self, record):
        msg = self.format(record)
        self.payload_handler.add_payload(
            {'type':'logentry',
             'log_entry':msg})

if os.path.exists(os.path.expanduser("~/.yt/favicon.ico")):
    ico = os.path.expanduser("~/.yt/")
else:
    ico = os.path.join(local_dir, "html", "resources", "images")
@route("/favicon.ico", method="GET")
def _favicon_ico():
    print ico
    return static_file("favicon.ico", ico)

class ExtProgressBar(object):
    def __init__(self, title, maxval):
        self.title = title
        self.maxval = maxval
        self.last = 0
        # Now we add a payload for the progress bar
        self.payload_handler = PayloadHandler()
        self.payload_handler.add_payload(
            {'type': 'widget',
             'widget_type': 'progressbar',
             'varname': 'pbar_top',
             'data': {'title':title}
            })

    def update(self, val):
        # An update is only meaningful if it's on the order of 1/100 or greater

        if (val - self.last) > (self.maxval / 100.0):
            self.last = val
            self.payload_handler.add_payload(
                {'type': 'widget_payload',
                 'widget_id': 'pbar_top',
                 'value': float(val) / self.maxval})

    def finish(self):
        self.payload_handler.add_payload(
            {'type': 'widget_payload',
             'widget_id': 'pbar_top',
             'value': -1})



"""
Author: Matthew Turk <matthewturk@gmail.com>
Affiliation:  UCSD
License:
  Copyright (C) 2010-2011 Matthew Turk  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# A simple image viewer, soon to be useful for stereo images, using OpenGL.

import sys, os
import OpenGL.GL as GL
import OpenGL.GLUT as GLUT
import OpenGL.GLU as GLU
import OpenGL.GL.shaders as shaders
from OpenGL.arrays import vbo, ArrayDatatype
import OpenGL.GL.ARB.framebuffer_object as GL_fbo
import Image
import glob
import numpy as np
import time

from small_apps import ViewHandler3D, GenericGLUTScene
from yt.visualization.image_writer import map_to_colors

from rendering_contexts import render_fbo, create_fbo, identity_view, \
        translate_view
from yt.visualization.volume_rendering.api import \
    HomogenizedVolume

ESCAPE = '\033'

_verts = ( (0,0,0), (1,0,0), (1,1,0), (0,1,0),
           (0,0,0), (1,0,0), (1,0,1), (0,0,1),
           (0,0,0), (0,0,1), (0,1,1), (0,1,0),
           (0,1,0), (1,1,0), (1,1,1), (0,1,1),
           (1,1,0), (1,0,0), (1,0,1), (1,1,1),
           (0,0,1), (0,1,1), (1,1,1), (1,0,1) )

_verts = ( (1,1,0), (0,1,0), (0,1,1), (1,1,1),
           (1,0,1), (0,0,1), (0,0,0), (1,0,0),
           (1,1,1), (0,1,1), (0,0,1), (1,0,1),
           (1,0,0), (0,0,0), (0,1,0), (1,1,0),
           (0,1,1), (0,1,0), (0,0,0), (0,0,1),
           (1,1,0), (1,1,1), (1,0,1), (1,0,0) )

_corner_list = [0,1,2,3, 4,5,6,7, 3,2,6,5, 0,4,7,1, 0,3,5,4, 1,7,6,2]
_corner_vals = [ (0,0,0), (1,0,0), (1,1,0), (0,1,0),
                 (0,0,1), (0,1,1), (1,1,1), (1,0,1) ]

_reversed = {2:0,1:1,0:2}

def _compress(lr, s, i):
    
    if lr == 0: return 0.5/s[_reversed[i]]
    else: return -0.5/s[_reversed[i]]
    

class MIPScene(GenericGLUTScene):
    _display_mode = (GLUT.GLUT_RGBA | GLUT.GLUT_DOUBLE | GLUT.GLUT_DEPTH)
    _title = "MIP"

    gl_state = None

    def _get_brick_vertices(self, offset):
        for b in self.hv.bricks:
            s = [ [-0.5, -0.5, -0.5],
                  [b.my_data[0].shape[i] - 1.5 for i in reversed(xrange(3))] ]
            for corner in _corner_list:
                for i, v in enumerate(_corner_vals[corner]):
                    yield s[v][i]

    def _get_texture_vertices(self):
        vs = [np.zeros(3, dtype='float32'),
              np.ones(3, dtype='float32')]
        #vs.reverse()
        for b in self.hv.bricks:
            shape = b.my_data[0].shape
            for corner in _corner_list:
                for i,v in enumerate(_corner_vals[corner]):
                    yield vs[v][i] + _compress(v, shape, i)

    def _setup_bricks(self):
        self._brick_textures = []
        for g in self.hv.bricks:
            self._upload_brick_textures(g)

    def _upload_brick_textures(self, brick):
        ix, iy, iz = brick.my_data[0].shape

        GL.glActiveTexture(GL.GL_TEXTURE0)
        id_field = GL.glGenTextures(1)
        upload = brick.my_data[0].astype("float32")
        #upload = (upload - -31.847) / ( -25.948 - -31.847 )
        #mi, ma = -31.847, -25.948
        #mi, ma = -27.2062, -20.9649 
        #upload = (upload - mi)/(ma - mi)
        self.mi = min(upload.min(), self.mi)
        self.ma = max(upload.max(), self.ma)
        #upload = (255*(upload - -31.0) / (-25.0 - -31.0)).astype("uint8")
        
        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glBindTexture(GL.GL_TEXTURE_3D, id_field)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_R, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexImage3D(GL.GL_TEXTURE_3D, 0, GL.GL_LUMINANCE32F_ARB, iz, iy, ix, 0,
                        GL.GL_LUMINANCE, GL.GL_FLOAT, upload)

        DW = self.hv.pf.domain_right_edge - self.hv.pf.domain_left_edge
        dds = ((brick.RightEdge - brick.LeftEdge) /
               (np.array([ix,iy,iz], dtype='float32')-1)) / DW
        BLE = brick.LeftEdge / DW - 0.5
        self._brick_textures.append(
            (id_field, (ix-1,iy-1,iz-1), dds, BLE))

        print "Uploaded", len(self._brick_textures)

    def _setup_colormap(self):

        buffer = np.mgrid[0.0:1.0:256j]
        colors = map_to_colors(buffer, "algae")
        
        GL.glActiveTexture(GL.GL_TEXTURE1)
        id_cmap = GL.glGenTextures(1)

        GL.glBindTexture(GL.GL_TEXTURE_1D, id_cmap)
        GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
        GL.glTexParameterf(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameterf(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR)
        GL.glTexParameterf(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR)
        GL.glTexImage1D(GL.GL_TEXTURE_1D, 0, GL.GL_RGBA, 256, 0,
                        GL.GL_RGBA, GL.GL_UNSIGNED_BYTE, colors)

        self.gl_state["cmap_tex"] = id_cmap

    def __init__(self, hv, offset = 0.5):
        self.offset = offset
        self.mi, self.ma = 1e30, -1e30
        self.hv = hv
        self.coord = 0.0
        self.tfac = 10.0
        self.rfac = 0.5
        self.wireframe = False
        self.glda = True
        self._setup_keypress_handler()
        self.gl_state = {}
        GenericGLUTScene.__init__(self, 800, 800)

        num = len(hv.bricks) * 6 * 4
        self.v = np.fromiter(self._get_brick_vertices(offset),
                             dtype = 'float32', count = num * 3)
        self.vertices = vbo.VBO(self.v)

        self.t = np.fromiter(self._get_texture_vertices(),
                             dtype = 'float32', count = num * 3)
        self.tvertices = vbo.VBO(self.t)

        self.ng = len(hv.bricks)
        self.position = np.zeros(3, dtype='float')
        self.rotation = np.zeros(3, dtype='float') + 30
        self.position[2] = -2 # Offset backwards a bit

        self._setup_bricks()
        create_fbo(self.gl_state)
        self._setup_colormap()

    def init_opengl(self, width, height):
        # One-time GL setup

        self.gl_state["width"] = width
        self.gl_state["height"] = height

        self.set_viewport()

        # Now we compile our shaders
        self.recompile()

    def set_viewport(self):
        if self.wireframe:
            GL.glClearColor(1, 1, 1, 1)
        else:
            GL.glClearColor(0, 0, 0, 1)
        GL.glColor3f(1, 0, 0)
        GL.glEnable(GL.GL_DEPTH_TEST)

        GL.glViewport(0, 0, self.gl_state["width"], self.gl_state["height"])
        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()
        GLU.gluPerspective(60., self.gl_state["width"] / float(self.gl_state["height"]), 1e-3, 10.)
        GL.glMatrixMode(GL.GL_MODELVIEW)

    def recompile(self):
        base = os.path.dirname(__file__) + "/"
        self.program = compileProgram(
            shaders.compileShader(
                open(base+"calculateRay.vertex.glsl").read(),
                GL.GL_VERTEX_SHADER),
            shaders.compileShader(
                open(base+"mip.fragment.glsl").read(),
                GL.GL_FRAGMENT_SHADER))

        def _set_simple_uniform(prog):
            GL.glUseProgram(prog)
            loc = GL.glGetUniformLocation(prog, 'buffer')
            GL.glUniform1i(loc, 0)
            loc = GL.glGetUniformLocation(prog, 'colormap')
            GL.glUniform1i(loc, 1)
            GL.glUseProgram(0)

        self.fbo_program = compileProgram(
            shaders.compileShader(
                open(base+"framebuffer.vertex.glsl").read(),
                GL.GL_VERTEX_SHADER),
            shaders.compileShader(
                open(base+"colormap.fragment.glsl").read(),
                GL.GL_FRAGMENT_SHADER),
            callback = _set_simple_uniform)

        self.uniform_locations = dict( (
                (v, GL.glGetUniformLocation(self.program, v)) for v in
                ['texture','shape','scaleBias','stepRatio','position']
            ) )
        self.fbo_uniform_locations = dict( (
                (v, GL.glGetUniformLocation(self.fbo_program, v)) for v in
                ['colormap','buffer']
            ) )

        print self.uniform_locations, self.fbo_uniform_locations

    @render_fbo
    def _draw_boxes(self):
        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        GL.glDisable(GL.GL_DEPTH_TEST)
        GL.glEnable(GL.GL_BLEND)
        if self.wireframe:
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE)
        else:
            GL.glBlendEquation(GL.GL_MAX)
            GL.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL)
            GL.glDisable(GL.GL_LINE_STIPPLE)
            GL.glDisable(GL.GL_LINE_SMOOTH)
            GL.glDisable(GL.GL_POINT_SMOOTH)
            GL.glEnable(GL.GL_CULL_FACE)
            #GL.glCullFace(GL.GL_FRONT)
            GL.glCullFace(GL.GL_BACK)

        GL.glLoadIdentity()
        GL.glTranslatef(*self.position)
        GL.glRotatef(self.rotation[0], 0, 0, 1)
        GL.glRotatef(self.rotation[1], 0, 1, 0)
        GL.glRotatef(self.rotation[2], 1, 0, 0)

        GL.glColor3f(0.0, 0.0, 0.0)
        GL.glUseProgram(self.program)
        scalebias = ( 1.0/(self.ma - self.mi), -self.mi)
        GL.glUniform1i(self.uniform_locations["texture"], 0)
        GL.glUniform2f(self.uniform_locations["scaleBias"], scalebias[0], scalebias[1])
        GL.glUniform1f(self.uniform_locations["stepRatio"], 1.0)
        GL.glUniform3f(self.uniform_locations["position"], *self.position)

        GL.glEnableClientState(GL.GL_VERTEX_ARRAY)
        GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        #GL.glEnableClientState(GL.GL_TEXTURE_COORD_ARRAY_EXT)

        GL.glColor3f(0.0, 0.0, 0.0)

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glEnable(GL.GL_TEXTURE_3D)

        self.tvertices.bind()
        GL.glTexCoordPointer(3, GL.GL_FLOAT, 0, self.tvertices)
        self.vertices.bind()
        GL.glVertexPointer(3, GL.GL_FLOAT, 0, self.vertices)

        for i, texinfo in enumerate(self._brick_textures):
            tex, shape, width, LE = texinfo
            GL.glActiveTexture(GL.GL_TEXTURE0)
            GL.glBindTexture(GL.GL_TEXTURE_3D, tex)
            GL.glUniform1f(self.uniform_locations["stepRatio"], 0.1)
            GL.glUniform3f(self.uniform_locations["shape"],
                shape[2], shape[1], shape[0])
            GL.glPushMatrix()
            GL.glTranslate(LE[2], LE[1], LE[0])
            GL.glScale(width[2], width[1], width[0])
            if self.glda:
                GL.glDrawArrays(GL.GL_QUADS, 24*i, 24)
            else:
                GL.glBegin(GL.GL_QUADS)
                for fi in range(6):
                    for v in range(4):
                        #print -LE, width
                        off = i*72+fi*12+v*3
                        qv = self.v[off:off+3]
                        tv = self.t[off:off+3]
                        GL.glTexCoord3f(*tv)
                        GL.glVertex3f(*qv)
                GL.glEnd()
            GL.glPopMatrix()
        GL.glDisableClientState(GL.GL_VERTEX_ARRAY)
        GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY)
        GL.glDisableClientState(GL.GL_TEXTURE_COORD_ARRAY_EXT)
        self.vertices.unbind()
        self.tvertices.unbind()
        GL.glDisable(GL.GL_CULL_FACE)

    def draw(self):

        self._draw_boxes()

        GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        GL.glUseProgram(self.fbo_program)

        GL.glMatrixMode(GL.GL_MODELVIEW)
        GL.glLoadIdentity()

        GL.glMatrixMode(GL.GL_PROJECTION)
        GL.glLoadIdentity()

        GL.glActiveTexture(GL.GL_TEXTURE0)
        GL.glEnable(GL.GL_TEXTURE_2D)
        GL.glBindTexture(GL.GL_TEXTURE_2D, self.gl_state["fbo_tex"])

        GL.glActiveTexture(GL.GL_TEXTURE1)
        GL.glEnable(GL.GL_TEXTURE_1D)
        GL.glBindTexture(GL.GL_TEXTURE_1D, self.gl_state["cmap_tex"])

        GL.glUniform1i(self.fbo_uniform_locations["buffer"], 0)
        GL.glUniform1i(self.fbo_uniform_locations["colormap"], 1)

        GL.glColor3f(0.3, 0.5, 1.0)

        GL.glBegin(GL.GL_QUADS)
        GL.glTexCoord2i(0, 0)
        GL.glVertex3i(-1, -1, -1)

        GL.glTexCoord2i(0, 1)
        GL.glVertex3i( 1, -1, -1)

        GL.glTexCoord2i(1, 1)
        GL.glVertex3i( 1,  1, -1)

        GL.glTexCoord2i(1, 0)
        GL.glVertex3i(-1,  1, -1)
        GL.glEnd()

        GL.glUseProgram(0)

        GLUT.glutSwapBuffers()

    def move_slice(self, value):
        self.coord += value

    def rotate(self, axis, value):
        self.rotation[axis] += value/self.rfac

    def reset_view(self):   
        print "RESETTING"
        self.position = np.zeros(3, dtype='float')
        self.rotation = np.zeros(3, dtype='float') + 30
        self.position[2] = -2 # Offset backwards a bit

    def translate(self, axis, value):
        self.position[axis] += value/self.tfac

    def toggle_wireframe(self):
        self.wireframe = not self.wireframe
        print "Wireframe:", self.wireframe
        
    def toggle_glda(self):
        self.glda = not self.glda
        print "GLDA:", self.glda
        
    def _setup_keypress_handler(self):
        self.keypress_handler = ViewHandler3D(self)
        self.keypress_handler.dispatch_table.update(dict(
            y = (self.move_slice, ( 0.05,)),
            h = (self.move_slice, (-0.05,)),
            t = (self.toggle_wireframe, ()),
            u = (self.toggle_glda, ()),
            i = (self.recompile, ()),
            o = (self.reset_view, ())
            ))

# We override the standard PyOpenGL one because otherwise we can't validate
# multitexture shaders.
def compileProgram(*my_shaders, **kwargs):
    """Create a new program, attach my_shaders and validate
    
    my_shaders -- arbitrary number of my_shaders to attach to the 
        generated program.
    
    This convenience function is *not* standard OpenGL,
    but it does wind up being fairly useful for demos 
    and the like.  You may wish to copy it to your code 
    base to guard against PyOpenGL changes.
    
    Usage:
    
        shader = compileProgram( 
            compileShader( source, GL_VERTEX_SHADER ),
            compileShader( source2, GL_FRAGMENT_SHADER ),
        )
        glUseProgram( shader )
    
    Note:
        If (and only if) validation of the linked program 
        *passes* then the passed-in shader objects will be 
        deleted from the GL.
    
    returns GLuint shader program reference
    raises RuntimeError when a link/validation failure occurs
    """
    program = GL.glCreateProgram()
    for shader in my_shaders:
        GL.glAttachShader(program, shader)
    GL.glLinkProgram(program)
    # Validation has to occur *after* linking
    if 'callback' in kwargs: kwargs['callback'](program)
    GL.glValidateProgram( program )
    validation = GL.glGetProgramiv( program, GL.GL_VALIDATE_STATUS )
    if validation == GL.GL_FALSE:
        raise RuntimeError(
            """Validation failure (%s): %s"""%(
            validation,
            GL.glGetProgramInfoLog( program ),
        ))
    link_status = GL.glGetProgramiv( program, GL.GL_LINK_STATUS )
    if link_status == GL.GL_FALSE:
        raise RuntimeError(
            """Link failure (%s): %s"""%(
            link_status,
            GL.glGetProgramInfoLog( program ),
        ))
    for shader in my_shaders:
        GL.glDeleteShader(shader)
    return shaders.ShaderProgram( program )

if __name__ == "__main__":
    import yt.convenience, yt.frontends.enzo.api
    pf = yt.convenience.load(sys.argv[-1])
    print pf
    hv = HomogenizedVolume(pf = pf, fields=["Density"], log_fields=[True])
    hv.initialize_source()
    mip = MIPScene(hv)
    mip.run()



"""
Answer Testing using Nose as a starting point

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import os
import hashlib
import contextlib
import urllib2
import cPickle
import sys
import cPickle
import shelve
import zlib

from matplotlib.testing.compare import compare_images
from nose.plugins import Plugin
from yt.testing import *
from yt.convenience import load, simulation
from yt.config import ytcfg
from yt.data_objects.static_output import StaticOutput
from yt.utilities.logger import disable_stream_logging
from yt.utilities.command_line import get_yt_version

import matplotlib.image as mpimg
import yt.visualization.plot_window as pw
import yt.utilities.progressbar as progressbar

mylog = logging.getLogger('nose.plugins.answer-testing')
run_big_data = False

# Set the latest gold and local standard filenames
_latest = ytcfg.get("yt", "gold_standard_filename")
_latest_local = ytcfg.get("yt", "local_standard_filename")
_url_path = "http://yt-answer-tests.s3-website-us-east-1.amazonaws.com/%s_%s"

class AnswerTesting(Plugin):
    name = "answer-testing"
    _my_version = None

    def options(self, parser, env=os.environ):
        super(AnswerTesting, self).options(parser, env=env)
        parser.add_option("--answer-name", dest="answer_name", metavar='str',
            default=None, help="The name of the standard to store/compare against")
        parser.add_option("--answer-store", dest="store_results", metavar='bool',
            default=False, action="store_true",
            help="Should we store this result instead of comparing?")
        parser.add_option("--local", dest="local_results",
            default=False, action="store_true", help="Store/load reference results locally?")
        parser.add_option("--answer-big-data", dest="big_data",
            default=False, help="Should we run against big data, too?",
            action="store_true")
        parser.add_option("--local-dir", dest="output_dir", metavar='str',
                          help="The name of the directory to store local results")

    @property
    def my_version(self, version=None):
        if self._my_version is not None:
            return self._my_version
        if version is None:
            try:
                version = get_yt_version()
            except:
                version = "UNKNOWN%s" % (time.time())
        self._my_version = version
        return self._my_version

    def configure(self, options, conf):
        super(AnswerTesting, self).configure(options, conf)
        if not self.enabled:
            return
        disable_stream_logging()

        # Parse through the storage flags to make sense of them
        # and use reasonable defaults
        # If we're storing the data, default storage name is local
        # latest version
        if options.store_results:
            if options.answer_name is None:
                self.store_name = _latest_local
            else:
                self.store_name = options.answer_name
            self.compare_name = None
        # if we're not storing, then we're comparing, and we want default
        # comparison name to be the latest gold standard
        # either on network or local
        else:
            if options.answer_name is None:
                if options.local_results:
                    self.compare_name = _latest_local
                else:
                    self.compare_name = _latest
            else:
                self.compare_name = options.answer_name
            self.store_name = self.my_version

        self.store_results = options.store_results

        ytcfg["yt","__withintesting"] = "True"
        AnswerTestingTest.result_storage = \
            self.result_storage = defaultdict(dict)
        if self.compare_name == "SKIP":
            self.compare_name = None
        elif self.compare_name == "latest":
            self.compare_name = _latest

        # Local/Cloud storage
        if options.local_results:
            if options.output_dir is None:
                print 'Please supply an output directory with the --local-dir option'
                sys.exit(1)
            storage_class = AnswerTestLocalStorage
            # Fix up filename for local storage
            if self.compare_name is not None:
                self.compare_name = "%s/%s/%s" % \
                    (os.path.realpath(options.output_dir), self.compare_name,
                     self.compare_name)
            if self.store_name is not None:
                name_dir_path = "%s/%s" % \
                    (os.path.realpath(options.output_dir),
                    self.store_name)
                if not os.path.isdir(name_dir_path):
                    os.makedirs(name_dir_path)
                self.store_name= "%s/%s" % \
                        (name_dir_path, self.store_name)
        else:
            storage_class = AnswerTestCloudStorage

        # Initialize answer/reference storage
        AnswerTestingTest.reference_storage = self.storage = \
                storage_class(self.compare_name, self.store_name)

        self.local_results = options.local_results
        global run_big_data
        run_big_data = options.big_data

    def finalize(self, result=None):
        if self.store_results is False: return
        self.storage.dump(self.result_storage)

    def help(self):
        return "yt answer testing support"

class AnswerTestStorage(object):
    def __init__(self, reference_name=None, answer_name=None):
        self.reference_name = reference_name
        self.answer_name = answer_name
        self.cache = {}
    def dump(self, result_storage, result):
        raise NotImplementedError
    def get(self, pf_name, default=None):
        raise NotImplementedError

class AnswerTestCloudStorage(AnswerTestStorage):
    def get(self, pf_name, default = None):
        if self.reference_name is None: return default
        if pf_name in self.cache: return self.cache[pf_name]
        url = _url_path % (self.reference_name, pf_name)
        try:
            resp = urllib2.urlopen(url)
        except urllib2.HTTPError as ex:
            raise YTNoOldAnswer(url)
        else:
            for this_try in range(3):
                try:
                    data = resp.read()
                except:
                    time.sleep(0.01)
                else:
                    # We were succesful
                    break
            else:
                # Raise error if all tries were unsuccessful
                raise YTCloudError(url)
            # This is dangerous, but we have a controlled S3 environment
            rv = cPickle.loads(data)
        self.cache[pf_name] = rv
        return rv

    def progress_callback(self, current, total):
        self.pbar.update(current)

    def dump(self, result_storage):
        if self.answer_name is None: return
        # This is where we dump our result storage up to Amazon, if we are able
        # to.
        import boto
        from boto.s3.key import Key
        c = boto.connect_s3()
        bucket = c.get_bucket("yt-answer-tests")
        for pf_name in result_storage:
            rs = cPickle.dumps(result_storage[pf_name])
            tk = bucket.get_key("%s_%s" % (self.answer_name, pf_name))
            if tk is not None: tk.delete()
            k = Key(bucket)
            k.key = "%s_%s" % (self.answer_name, pf_name)

            pb_widgets = [
                unicode(k.key, errors='ignore').encode('utf-8'), ' ',
                progressbar.FileTransferSpeed(),' <<<', progressbar.Bar(),
                '>>> ', progressbar.Percentage(), ' ', progressbar.ETA()
                ]
            self.pbar = progressbar.ProgressBar(widgets=pb_widgets,
                                                maxval=sys.getsizeof(rs))

            self.pbar.start()
            k.set_contents_from_string(rs, cb=self.progress_callback,
                                       num_cb=100000)
            k.set_acl("public-read")
            self.pbar.finish()

class AnswerTestLocalStorage(AnswerTestStorage):
    def dump(self, result_storage):
        if self.answer_name is None: return
        # Store data using shelve
        ds = shelve.open(self.answer_name, protocol=-1)
        for pf_name in result_storage:
            answer_name = "%s" % pf_name
            if answer_name in ds:
                mylog.info("Overwriting %s", answer_name)
            ds[answer_name] = result_storage[pf_name]
        ds.close()

    def get(self, pf_name, default=None):
        if self.reference_name is None: return default
        # Read data using shelve
        answer_name = "%s" % pf_name
        ds = shelve.open(self.reference_name, protocol=-1)
        try:
            result = ds[answer_name]
        except KeyError:
            result = default
        ds.close()
        return result

@contextlib.contextmanager
def temp_cwd(cwd):
    oldcwd = os.getcwd()
    os.chdir(cwd)
    yield
    os.chdir(oldcwd)

def can_run_pf(pf_fn):
    if isinstance(pf_fn, StaticOutput):
        return AnswerTestingTest.result_storage is not None
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    with temp_cwd(path):
        try:
            load(pf_fn)
        except:
            return False
    return AnswerTestingTest.result_storage is not None

def data_dir_load(pf_fn):
    path = ytcfg.get("yt", "test_data_dir")
    if isinstance(pf_fn, StaticOutput): return pf_fn
    if not os.path.isdir(path):
        return False
    with temp_cwd(path):
        pf = load(pf_fn)
        pf.h
        return pf

def sim_dir_load(sim_fn, path = None, sim_type = "Enzo",
                 find_outputs=False):
    if path is None and not os.path.exists(sim_fn):
        raise IOError
    if os.path.exists(sim_fn) or not path:
        path = "."
    with temp_cwd(path):
        return simulation(sim_fn, sim_type,
                          find_outputs=find_outputs)

class AnswerTestingTest(object):
    reference_storage = None
    result_storage = None
    prefix = ""
    def __init__(self, pf_fn):
        self.pf = data_dir_load(pf_fn)

    def __call__(self):
        nv = self.run()
        if self.reference_storage.reference_name is not None:
            dd = self.reference_storage.get(self.storage_name)
            if dd is None or self.description not in dd:
                raise YTNoOldAnswer("%s : %s" % (self.storage_name , self.description))
            ov = dd[self.description]
            self.compare(nv, ov)
        else:
            ov = None
        self.result_storage[self.storage_name][self.description] = nv

    @property
    def storage_name(self):
        if self.prefix != "":
            return "%s_%s" % (self.prefix, self.pf)
        return str(self.pf)

    def compare(self, new_result, old_result):
        raise RuntimeError

    def create_obj(self, pf, obj_type):
        # obj_type should be tuple of
        #  ( obj_name, ( args ) )
        if obj_type is None:
            return pf.h.all_data()
        cls = getattr(pf.h, obj_type[0])
        obj = cls(*obj_type[1])
        return obj

    def create_plot(self, pf, plot_type, plot_field, plot_axis, plot_kwargs = None):
        # plot_type should be a string
        # plot_args should be a tuple
        # plot_kwargs should be a dict
        if plot_type is None:
            raise RuntimeError('Must explicitly request a plot type')
        cls = getattr(pw, plot_type)
        plot = cls(*(pf, plot_axis, plot_field), **plot_kwargs)
        return plot

    @property
    def sim_center(self):
        """
        This returns the center of the domain.
        """
        return 0.5*(self.pf.domain_right_edge + self.pf.domain_left_edge)

    @property
    def max_dens_location(self):
        """
        This is a helper function to return the location of the most dense
        point.
        """
        return self.pf.h.find_max("Density")[1]

    @property
    def entire_simulation(self):
        """
        Return an unsorted array of values that cover the entire domain.
        """
        return self.pf.h.all_data()

    @property
    def description(self):
        obj_type = getattr(self, "obj_type", None)
        if obj_type is None:
            oname = "all"
        else:
            oname = "_".join((str(s) for s in obj_type))
        args = [self._type_name, str(self.pf), oname]
        args += [str(getattr(self, an)) for an in self._attrs]
        return "_".join(args)

class FieldValuesTest(AnswerTestingTest):
    _type_name = "FieldValues"
    _attrs = ("field", )

    def __init__(self, pf_fn, field, obj_type = None,
                 decimals = None):
        super(FieldValuesTest, self).__init__(pf_fn)
        self.obj_type = obj_type
        self.field = field
        self.decimals = decimals

    def run(self):
        obj = self.create_obj(self.pf, self.obj_type)
        avg = obj.quantities["WeightedAverageQuantity"](self.field,
                             weight="Ones")
        (mi, ma), = obj.quantities["Extrema"](self.field)
        return np.array([avg, mi, ma])

    def compare(self, new_result, old_result):
        err_msg = "Field values for %s not equal." % self.field
        if self.decimals is None:
            assert_equal(new_result, old_result,
                         err_msg=err_msg, verbose=True)
        else:
            assert_allclose(new_result, old_result, 10.**(-self.decimals),
                             err_msg=err_msg, verbose=True)

class AllFieldValuesTest(AnswerTestingTest):
    _type_name = "AllFieldValues"
    _attrs = ("field", )

    def __init__(self, pf_fn, field, obj_type = None,
                 decimals = None):
        super(AllFieldValuesTest, self).__init__(pf_fn)
        self.obj_type = obj_type
        self.field = field
        self.decimals = decimals

    def run(self):
        obj = self.create_obj(self.pf, self.obj_type)
        return obj[self.field]

    def compare(self, new_result, old_result):
        err_msg = "All field values for %s not equal." % self.field
        if self.decimals is None:
            assert_equal(new_result, old_result,
                         err_msg=err_msg, verbose=True)
        else:
            assert_rel_equal(new_result, old_result, self.decimals,
                             err_msg=err_msg, verbose=True)

class ProjectionValuesTest(AnswerTestingTest):
    _type_name = "ProjectionValues"
    _attrs = ("field", "axis", "weight_field")

    def __init__(self, pf_fn, axis, field, weight_field = None,
                 obj_type = None, decimals = None):
        super(ProjectionValuesTest, self).__init__(pf_fn)
        self.axis = axis
        self.field = field
        self.weight_field = weight_field
        self.obj_type = obj_type
        self.decimals = decimals

    def run(self):
        if self.obj_type is not None:
            obj = self.create_obj(self.pf, self.obj_type)
        else:
            obj = None
        if self.pf.domain_dimensions[self.axis] == 1: return None
        proj = self.pf.h.proj(self.axis, self.field,
                              weight_field=self.weight_field,
                              data_source = obj)
        return proj.field_data

    def compare(self, new_result, old_result):
        if new_result is None:
            return
        assert(len(new_result) == len(old_result))
        for k in new_result:
            assert (k in old_result)
        for k in new_result:
            err_msg = "%s values of %s (%s weighted) projection (axis %s) not equal." % \
              (k, self.field, self.weight_field, self.axis)
            if k == 'weight_field' and self.weight_field is None:
                continue
            if self.decimals is None:
                assert_equal(new_result[k], old_result[k],
                             err_msg=err_msg)
            else:
                assert_allclose(new_result[k], old_result[k],
                                 10.**-(self.decimals), err_msg=err_msg)

class PixelizedProjectionValuesTest(AnswerTestingTest):
    _type_name = "PixelizedProjectionValues"
    _attrs = ("field", "axis", "weight_field")

    def __init__(self, pf_fn, axis, field, weight_field = None,
                 obj_type = None):
        super(PixelizedProjectionValuesTest, self).__init__(pf_fn)
        self.axis = axis
        self.field = field
        self.weight_field = field
        self.obj_type = obj_type

    def run(self):
        if self.obj_type is not None:
            obj = self.create_obj(self.pf, self.obj_type)
        else:
            obj = None
        proj = self.pf.h.proj(self.axis, self.field,
                              weight_field=self.weight_field,
                              data_source = obj)
        frb = proj.to_frb((1.0, 'unitary'), 256)
        frb[self.field]
        frb[self.weight_field]
        d = frb.data
        d.update( dict( (("%s_sum" % f, proj[f].sum(dtype="float64"))
                         for f in proj.field_data.keys()) ) )
        return d

    def compare(self, new_result, old_result):
        assert(len(new_result) == len(old_result))
        for k in new_result:
            assert (k in old_result)
        for k in new_result:
            assert_rel_equal(new_result[k], old_result[k], 10)

class GridValuesTest(AnswerTestingTest):
    _type_name = "GridValues"
    _attrs = ("field",)

    def __init__(self, pf_fn, field):
        super(GridValuesTest, self).__init__(pf_fn)
        self.field = field

    def run(self):
        hashes = {}
        for g in self.pf.h.grids:
            hashes[g.id] = hashlib.md5(g[self.field].tostring()).hexdigest()
            g.clear_data()
        return hashes

    def compare(self, new_result, old_result):
        assert(len(new_result) == len(old_result))
        for k in new_result:
            assert (k in old_result)
        for k in new_result:
            assert_equal(new_result[k], old_result[k])

class VerifySimulationSameTest(AnswerTestingTest):
    _type_name = "VerifySimulationSame"
    _attrs = ()

    def __init__(self, simulation_obj):
        self.pf = simulation_obj

    def run(self):
        result = [ds.current_time for ds in self.pf]
        return result

    def compare(self, new_result, old_result):
        assert_equal(len(new_result), len(old_result),
                     err_msg="Number of outputs not equal.",
                     verbose=True)
        for i in range(len(new_result)):
            assert_equal(new_result[i], old_result[i],
                         err_msg="Output times not equal.",
                         verbose=True)

class GridHierarchyTest(AnswerTestingTest):
    _type_name = "GridHierarchy"
    _attrs = ()

    def run(self):
        result = {}
        result["grid_dimensions"] = self.pf.h.grid_dimensions
        result["grid_left_edges"] = self.pf.h.grid_left_edge
        result["grid_right_edges"] = self.pf.h.grid_right_edge
        result["grid_levels"] = self.pf.h.grid_levels
        result["grid_particle_count"] = self.pf.h.grid_particle_count
        return result

    def compare(self, new_result, old_result):
        for k in new_result:
            assert_equal(new_result[k], old_result[k])

class ParentageRelationshipsTest(AnswerTestingTest):
    _type_name = "ParentageRelationships"
    _attrs = ()
    def run(self):
        result = {}
        result["parents"] = []
        result["children"] = []
        for g in self.pf.h.grids:
            p = g.Parent
            if p is None:
                result["parents"].append(None)
            elif hasattr(p, "id"):
                result["parents"].append(p.id)
            else:
                result["parents"].append([pg.id for pg in p])
            result["children"].append([c.id for c in g.Children])
        return result

    def compare(self, new_result, old_result):
        for newp, oldp in zip(new_result["parents"], old_result["parents"]):
            assert(newp == oldp)
        for newc, oldc in zip(new_result["children"], old_result["children"]):
            assert(newp == oldp)

class PlotWindowAttributeTest(AnswerTestingTest):
    _type_name = "PlotWindowAttribute"
    _attrs = ('plot_type', 'plot_field', 'plot_axis', 'attr_name', 'attr_args')
    def __init__(self, pf_fn, plot_field, plot_axis, attr_name, attr_args,
                 decimals, plot_type = 'SlicePlot'):
        super(PlotWindowAttributeTest, self).__init__(pf_fn)
        self.plot_type = plot_type
        self.plot_field = plot_field
        self.plot_axis = plot_axis
        self.plot_kwargs = {}
        self.attr_name = attr_name
        self.attr_args = attr_args
        self.decimals = decimals

    def run(self):
        plot = self.create_plot(self.pf, self.plot_type, self.plot_field,
                                self.plot_axis, self.plot_kwargs)
        attr = getattr(plot, self.attr_name)
        attr(*self.attr_args[0], **self.attr_args[1])
        fn = plot.save()[0]
        image = mpimg.imread(fn)
        os.remove(fn)
        return [zlib.compress(image.dumps())]

    def compare(self, new_result, old_result):
        fns = ['old.png', 'new.png']
        mpimg.imsave(fns[0], np.loads(zlib.decompress(old_result[0])))
        mpimg.imsave(fns[1], np.loads(zlib.decompress(new_result[0])))
        assert compare_images(fns[0], fns[1], 10**(-self.decimals)) == None
        for fn in fns: os.remove(fn)

def requires_pf(pf_fn, big_data = False):
    def ffalse(func):
        return lambda: None
    def ftrue(func):
        return func
    if run_big_data == False and big_data == True:
        return ffalse
    elif not can_run_pf(pf_fn):
        return ffalse
    else:
        return ftrue

def small_patch_amr(pf_fn, fields):
    if not can_run_pf(pf_fn): return
    dso = [ None, ("sphere", ("max", (0.1, 'unitary')))]
    yield GridHierarchyTest(pf_fn)
    yield ParentageRelationshipsTest(pf_fn)
    for field in fields:
        yield GridValuesTest(pf_fn, field)
        for axis in [0, 1, 2]:
            for ds in dso:
                for weight_field in [None, "Density"]:
                    yield ProjectionValuesTest(
                        pf_fn, axis, field, weight_field,
                        ds)
                yield FieldValuesTest(
                        pf_fn, field, ds)

def big_patch_amr(pf_fn, fields):
    if not can_run_pf(pf_fn): return
    dso = [ None, ("sphere", ("max", (0.1, 'unitary')))]
    yield GridHierarchyTest(pf_fn)
    yield ParentageRelationshipsTest(pf_fn)
    for field in fields:
        yield GridValuesTest(pf_fn, field)
        for axis in [0, 1, 2]:
            for ds in dso:
                for weight_field in [None, "Density"]:
                    yield PixelizedProjectionValuesTest(
                        pf_fn, axis, field, weight_field,
                        ds)

class AssertWrapper(object):
    """
    Used to wrap a numpy testing assertion, in order to provide a useful name
    for a given assertion test.
    """
    def __init__(self, description, *args):
        # The key here is to add a description attribute, which nose will pick
        # up.
        self.args = args
        self.description = description

    def __call__(self):
        self.args[0](*self.args[1:])



"""
A light interface to FreeType2

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: UCSD
Homepage: http://yt-project.org/
License:
  Copyright (C) 2010-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython

cdef int initialized = 0

from libc.stdio cimport fopen, fclose, FILE

cdef extern from "freetype_includes.h":
    # NOTE that size_t might not be int
    ctypedef struct FT_Library:
        pass

    ctypedef struct FT_Bitmap:
        unsigned char *buffer
        int width
        int rows

    ctypedef struct FT_Vector:
        int x
        int y

    ctypedef struct FT_GlyphSlotRec:
        FT_Bitmap bitmap
        int bitmap_left
        int bitmap_top
        FT_Vector advance
    ctypedef FT_GlyphSlotRec* FT_GlyphSlot

    ctypedef struct FT_FaceRec:
        FT_GlyphSlot glyph
    ctypedef FT_FaceRec* FT_Face

    int FT_LOAD_RENDER

    int FT_Init_FreeType( FT_Library *lib)
    int FT_New_Face(FT_Library library, char* filepathname, long face_index,
                    FT_Face *aface)
    int FT_Set_Char_Size(FT_Face face, double width, double height,
                         unsigned dpi_w, unsigned dpi_h)
    int FT_Load_Char(FT_Face face, unsigned long char_code, int load_flags)

cdef FT_Library library

def initialize_library():
    # We do not clear the library from memory.
    global initialized
    if initialized == 1: return
    cdef int error = FT_Init_FreeType(&library)
    if error: raise RuntimeError
    initialized = 1

def simple_writing(char *fontpath, long face_index, long dpi,
                   long font_size, char *text,
                   np.ndarray[np.uint8_t, ndim=3] buffer,
                   int xpos, int ypos):
    initialize_library()
    cdef FT_Face face
    cdef int error = FT_New_Face(library, fontpath, face_index, &face)
    if error: raise RuntimeError(error)
    error = FT_Set_Char_Size(face, 0, font_size * 64, dpi, dpi)
    if error: raise RuntimeError(error)
    cdef FT_GlyphSlot slot = face.glyph
    cdef int n, i, j, c
    cdef pi = xpos, pj = ypos # Pen locations
    cdef di, dj # Where we actually draw
    cdef char tchar
    cdef int ichar
    cdef np.uint8_t bb
    for n in range(len(text)):
        tchar = text[n]
        ichar = <int> tchar
        error = FT_Load_Char(face, ichar, FT_LOAD_RENDER)
        if error: raise RuntimeError(error)
        for i in range(slot.bitmap.width):
            for j in range(slot.bitmap.rows):
                di = pi + slot.bitmap_left + i
                dj = pj - slot.bitmap_top + j
                for c in range(3):
                    bb = <np.uint8_t> slot.bitmap.buffer[j * slot.bitmap.width + i]
                    buffer[dj,di,c] |= bb
        pi += slot.advance.x / 64
        pj += slot.advance.y / 64



#!/usr/bin/env python
import os
import sys
import os.path
import glob


# snatched from PyTables
def add_from_path(envname, dirs):
    try:
        dirs.extend(os.environ[envname].split(os.pathsep))
    except KeyError:
        pass


# snatched from PyTables
def add_from_flags(envname, flag_key, dirs):
    for flag in os.environ.get(envname, "").split():
        if flag.startswith(flag_key):
            dirs.append(flag[len(flag_key):])


# snatched from PyTables
def get_default_dirs():
    default_header_dirs = []
    add_from_path("CPATH", default_header_dirs)
    add_from_path("C_INCLUDE_PATH", default_header_dirs)
    add_from_flags("CPPFLAGS", "-I", default_header_dirs)
    default_header_dirs.extend(
        ['/usr/include', '/usr/local/include', '/usr/X11']
    )

    default_library_dirs = []
    add_from_flags("LDFLAGS", "-L", default_library_dirs)
    default_library_dirs.extend(
        os.path.join(_tree, _arch)
        for _tree in ('/', '/usr', '/usr/local', '/usr/X11')
        for _arch in ('lib64', 'lib')
    )
    return default_header_dirs, default_library_dirs


def get_location_from_env(env):
    env_dir = os.environ[env]
    env_inc = os.path.join(env_dir, "include")
    env_lib = os.path.join(env_dir, "lib")
    print("%s_LOCATION: %s: %s, %s"
          % (env.split('_')[0], env, env_inc, env_lib))
    return (env_inc, env_lib)


def get_location_from_cfg(cfg):
    cfg_dir = open(cfg).read().strip()
    cfg_inc = os.path.join(cfg_dir, "include")
    cfg_lib = os.path.join(cfg_dir, "lib")
    print("%s_LOCATION: %s: %s, %s"
          % (cfg.split('.')[0].upper(), cfg, cfg_inc, cfg_lib))
    return (cfg_inc, cfg_lib)


def check_prefix(inc_dir, lib_dir):
    prefix = os.path.commonprefix([inc_dir, lib_dir]).rstrip('/\\')
    if prefix is not '' and prefix == os.path.dirname(inc_dir):
        return (inc_dir, lib_dir)
    else:
        print("It seems that include prefix is different from lib prefix")
        print("Please use either env variable or cfg to set proper path")
        return (None, None)


def get_location_from_ctypes(header, library):
    try:
        import ctypes
        import ctypes.util
    except ImportError:
        return (None, None)

    default_header_dirs, default_library_dirs = get_default_dirs()
    target_inc, target_libdir = None, None
    for inc_prefix in default_header_dirs:
        if os.path.isfile(os.path.join(inc_prefix, header)):
            target_inc = inc_prefix

    target_libfile = ctypes.util.find_library(library)
    if target_libfile is not None and os.path.isfile(target_libfile):
        return check_prefix(target_inc, os.path.dirname(target_libfile))
    for lib_dir in default_library_dirs:
        try:
            ctypes.CDLL(os.path.join(lib_dir, target_libfile))
            target_libdir = lib_dir
        except OSError:
            pass
    return check_prefix(target_inc, target_libdir)


def check_for_dependencies(env, cfg, header, library):
    # First up: check in environment
    if env in os.environ:
        return get_location_from_env(env)
    # Next up, we try config file
    elif os.path.exists(cfg):
        return get_location_from_cfg(cfg)
    # Now we see if ctypes can help us
    if os.name == 'posix':
        target_inc, target_lib = get_location_from_ctypes(header, library)
    if None not in (target_inc, target_lib):
        print(
            "%s_LOCATION: %s found via ctypes in: %s, %s"
            % (env.split('_')[0], env.split('_')[0], target_inc, target_lib)
        )
        return (target_inc, target_lib)

    print("Reading %s location from %s failed." % (env.split('_')[0], cfg))
    print("Please place the base directory of your")
    print("%s install in %s and restart." % (env.split('_')[0], cfg))
    print("(ex: \"echo '/usr/local/' > %s\" )" % cfg)
    print("You can locate the path by looking for %s" % header)
    sys.exit(1)


def check_for_hdf5():
    return check_for_dependencies("HDF5_DIR", "hdf5.cfg", "hdf5.h", "hdf5")


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    config = Configuration('utilities', parent_package, top_path)
    config.add_subpackage("amr_kdtree")
    config.add_subpackage("poster")
    config.add_subpackage("answer_testing")
    config.add_subpackage("delaunay")  # From SciPy, written by Robert Kern
    config.add_subpackage("kdtree")
    config.add_subpackage("spatial")
    config.add_subpackage("grid_data_format")
    config.add_subpackage("parallel_tools")
    config.add_subpackage("lib")
    config.add_extension("data_point_utilities",
                         "yt/utilities/data_point_utilities.c",
                         libraries=["m"])
    config.add_subpackage("tests")
    hdf5_inc, hdf5_lib = check_for_hdf5()
    include_dirs = [hdf5_inc]
    library_dirs = [hdf5_lib]
    config.add_extension("hdf5_light_reader",
                         "yt/utilities/hdf5_light_reader.c",
                         define_macros=[("H5_USE_16_API", True)],
                         libraries=["m", "hdf5"],
                         library_dirs=library_dirs, include_dirs=include_dirs)
    config.add_extension("libconfig_wrapper",
                         ["yt/utilities/libconfig_wrapper.pyx"] +
                         glob.glob("yt/utilities/_libconfig/*.c"),
                         include_dirs=["yt/utilities/_libconfig/"],
                         define_macros=[("HAVE_XLOCALE_H", True)]
                         )
    config.make_config_py()  # installs __config__.py
    # config.make_svn_version_py()
    return config



"""
Athena-specific fields

Author: Samuel W. Skillman <samskillman@gmail.com>
Affiliation: University of Colorado at Boulder
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Author: John A. ZuHone <jzuhone@gmail.com>
Affiliation: NASA/Goddard Space Flight Center
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Samuel W. Skillman, Matthew Turk, J. S. Oishi.  
  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    FieldInfo, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType, \
    NullFunc, \
    TranslationFunc
from yt.utilities.physical_constants import \
    kboltz,mh
import yt.data_objects.universal_fields

AthenaFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = AthenaFieldInfo.add_field

KnownAthenaFields = FieldInfoContainer()
add_athena_field = KnownAthenaFields.add_field

add_athena_field("density", function=NullFunc, take_log=False,
                 units=r"", projected_units =r"")

add_athena_field("pressure", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("total_energy", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("velocity_x", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("velocity_y", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("velocity_z", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("momentum_x", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("momentum_y", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("momentum_z", function=NullFunc, take_log=False,
                 units=r"")

add_athena_field("cell_centered_B_x", function=NullFunc, take_log=False,
                 units=r"", display_name=r"$\rm{cell\/centered\/B_x}$")

add_athena_field("cell_centered_B_y", function=NullFunc, take_log=False,
                 units=r"", display_name=r"$\rm{cell\/centered\/B_y}$")

add_athena_field("cell_centered_B_z", function=NullFunc, take_log=False,
                 units=r"", display_name=r"$\rm{cell\/centered\/B_z}$")

# In Athena, conservative or primitive variables may be written out.
# By default, yt concerns itself with primitive variables. The following
# field definitions allow for conversions to primitive variables in the
# case that the file contains the conservative ones.

def _convertDensity(data) :
    return data.convert("Density")
def _density(field, data) :
    return data["density"]
add_field("Density", function=_density, take_log=False,
          units=r"\rm{g}/\rm{cm}^3", projected_units=r"\rm{g}/\rm{cm}^2",
          convert_function=_convertDensity)

def _convertVelocity(data):
    return data.convert("x-velocity")
def _xvelocity(field, data):
    if "velocity_x" in data.pf.field_info:
        return data["velocity_x"]
    else:
        return data["momentum_x"]/data["density"]           
add_field("x-velocity", function=_xvelocity, take_log=False,
          units=r"\rm{cm}/\rm{s}", convert_function=_convertVelocity)
def _yvelocity(field, data):
    if "velocity_y" in data.pf.field_info:
        return data["velocity_y"]
    else:
        return data["momentum_y"]/data["density"]
add_field("y-velocity", function=_yvelocity, take_log=False,
          units=r"\rm{cm}/\rm{s}", convert_function=_convertVelocity)
def _zvelocity(field, data):
    if "velocity_z" in data.pf.field_info:
        return data["velocity_z"]
    else:
        return data["momentum_z"]/data["density"]
add_field("z-velocity", function=_zvelocity, take_log=False,
          units=r"\rm{cm}/\rm{s}", convert_function=_convertVelocity)

def _convertEnergy(data) :
    return data.convert("x-velocity")**2
def _gasenergy(field, data) :
    if "pressure" in data.pf.field_info:
        return data["pressure"]/(data.pf["Gamma"]-1.0)/data["density"]
    else:
        return (data["total_energy"] - 
                0.5*(data["cell_centered_B_x"]**2 +
                     data["cell_centered_B_y"]**2 +
                     data["cell_centered_B_z"]**2) - 
                0.5*(data["momentum_x"]**2 +
                     data["momentum_y"]**2 +
                     data["momentum_z"]**2)/data["density"])/data["density"]
add_field("Gas_Energy", function=_gasenergy, take_log=False,
          units=r"\rm{erg}/\rm{g}")

def _convertPressure(data) :
    return data.convert("Density")*data.convert("x-velocity")**2
def _pressure(field, data) :
    if "pressure" in data.pf.field_info:
        return data["pressure"]
    else:
        return (data["total_energy"] -
                0.5*(data["cell_centered_B_x"]**2 +
                     data["cell_centered_B_y"]**2 +
                     data["cell_centered_B_z"]**2) -
                0.5*(data["momentum_x"]**2 +
                     data["momentum_y"]**2 +
                     data["momentum_z"]**2)/data["density"])*(data.pf["Gamma"]-1.0)
add_field("Pressure", function=_pressure, take_log=False, convert_function=_convertPressure,
          units=r"\rm{erg}/\rm{cm}^3", projected_units=r"\rm{erg}/\rm{cm}^2")

def _temperature(field, data):
    if data.has_field_parameter("mu"):
        mu = data.get_field_parameter("mu")
    else:
        mu = 0.6
    return mu*mh*data["Pressure"]/data["Density"]/kboltz
add_field("Temperature", function=_temperature, take_log=False,
          units=r"\rm{K}")

def _convertBfield(data):
        return np.sqrt(4*np.pi*data.convert("Density")*data.convert("x-velocity")**2)
def _Bx(field, data):
    return data['cell_centered_B_x']
add_field("Bx", function=_Bx, take_log=False,
          units=r"\rm{Gauss}", display_name=r"B_x",
          convert_function=_convertBfield)
def _By(field, data):
    return data['cell_centered_B_y']
add_field("By", function=_By, take_log=False,
          units=r"\rm{Gauss}", display_name=r"B_y",
          convert_function=_convertBfield)
def _Bz(field, data):
    return data['cell_centered_B_z']
add_field("Bz", function=_Bz, take_log=False,
          units=r"\rm{Gauss}", display_name=r"B_z",
          convert_function=_convertBfield)





"""
Simple integrators for the radiative transfer equation

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free

cdef extern from "math.h":
    double exp(double x) nogil
    float expf(float x) nogil
    long double expl(long double x) nogil
    double floor(double x) nogil
    double ceil(double x) nogil
    double fmod(double x, double y) nogil
    double log2(double x) nogil
    long int lrint(double x) nogil
    double fabs(double x) nogil

# These routines are separated into a couple different categories:
#
#   * Routines for identifying intersections of an object with a bounding box
#   * Routines for identifying cells/points inside a bounding box that
#     intersect with an object
#   * Routines that speed up some type of geometric calculation

# First, bounding box / object intersection routines.
# These all respect the interface "dobj" and a set of left_edges, right_edges,
# sometimes also accepting level and mask information.

def ortho_ray_grids(dobj, np.ndarray[np.float64_t, ndim=2] left_edges,
                          np.ndarray[np.float64_t, ndim=2] right_edges):
    cdef int i
    cdef int ng = left_edges.shape[0]
    cdef int px_ax = dobj.px_ax
    cdef int py_ax = dobj.py_ax
    cdef np.float64_t px = dobj.px
    cdef np.float64_t py = dobj.py
    cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
    for i in range(ng):
        if (    (px >= left_edges[i, px])
            and (px < right_edges[i, px])
            and (py >= left_edges[i, py])
            and (py < right_edges[i, py])):
            gridi[i] = 1
    return gridi

def ray_grids(dobj, np.ndarray[np.float64_t, ndim=2] left_edges,
                    np.ndarray[np.float64_t, ndim=2] right_edges):
    cdef int i, ax
    cdef int i1, i2
    cdef int ng = left_edges.shape[0]
    cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
    cdef np.float64_t vs[3], t, p0[3], p1[3], v[3]
    for i in range(3):
        p0[i] = dobj.start_point[i]
        p1[i] = dobj.end_point[i]
        v[i] = dobj.vec[i]
    # We check first to see if at any point, the ray intersects a grid face
    for gi in range(ng):
        for ax in range(3):
            i1 = (ax+1) % 3
            i2 = (ax+2) % 3
            t = (left_edges[gi,ax] - p0[ax])/v[ax]
            for i in range(3):
                vs[i] = t * v[i] + p0[i]
            if left_edges[gi,i1] <= vs[i1] and \
               right_edges[gi,i1] >= vs[i1] and \
               left_edges[gi,i2] <= vs[i2] and \
               right_edges[gi,i2] >= vs[i2]:
                gridi[gi] = 1
                break
            t = (right_edges[gi,ax] - p0[ax])/v[ax]
            for i in range(3):
                vs[i] = t * v[i] + p0[i]
            if left_edges[gi,i1] <= vs[i1] and \
               right_edges[gi,i1] >= vs[i1] and \
               left_edges[gi,i2] <= vs[i2] and \
               right_edges[gi,i2] >= vs[i2]:
                gridi[gi] = 1
                break
        if gridi[gi] == 1: continue
        # if the point is fully enclosed, we count the grid
        if left_edges[gi,0] <= p0[0] and \
           right_edges[gi,0] >= p0[0] and \
           left_edges[gi,1] <= p0[1] and \
           right_edges[gi,1] >= p0[1] and \
           left_edges[gi,2] <= p0[2] and \
           right_edges[gi,2] >= p0[2]:
            gridi[gi] = 1
            continue
        if left_edges[gi,0] <= p1[0] and \
           right_edges[gi,0] >= p1[0] and \
           left_edges[gi,1] <= p1[1] and \
           right_edges[gi,1] >= p1[1] and \
           left_edges[gi,2] <= p1[2] and \
           right_edges[gi,2] >= p1[2]:
            gridi[gi] = 1
            continue
    return gridi

def slice_grids(dobj, np.ndarray[np.float64_t, ndim=2] left_edges,
                      np.ndarray[np.float64_t, ndim=2] right_edges):
    cdef int i, ax
    cdef int ng = left_edges.shape[0]
    cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
    ax = dobj.axis
    cdef np.float64_t coord = dobj.coord
    for i in range(ng):
        if left_edges[i, ax] <= coord and \
           right_edges[i, ax] > coord:
            gridi[i] = 1
    return gridi

def cutting_plane_grids(dobj, np.ndarray[np.float64_t, ndim=2] left_edges,
                        np.ndarray[np.float64_t, ndim=2] right_edges):
    cdef int i
    cdef int ng = left_edges.shape[0]
    cdef np.ndarray[np.uint8_t, ndim=1] gridi = np.zeros(ng, dtype='uint8')
    cdef np.float64_t *arr[2]
    arr[0] = <np.float64_t *> left_edges.data
    arr[1] = <np.float64_t *> right_edges.data
    cdef np.float64_t x, y, z
    cdef np.float64_t norm_vec[3]
    cdef np.float64_t d = dobj._d # offset to center
    cdef np.float64_t gd # offset to center
    cdef np.int64_t all_under, all_over
    for i in range(3):
        norm_vec[i] = dobj._norm_vec[i]
    for i in range(ng):
        all_under = 1
        all_over = 1
        # Check each corner
        for xi in range(2):
            x = arr[xi][i * 3 + 0]
            for yi in range(2):
                y = arr[yi][i * 3 + 1]
                for zi in range(2):
                    z = arr[zi][i * 3 + 2]
                    gd = ( x*norm_vec[0]
                         + y*norm_vec[1]
                         + z*norm_vec[2]) + d
                    if gd <= 0: all_over = 0
                    if gd >= 0: all_under = 0
        if not (all_over == 1 or all_under == 1):
            gridi[i] = 1
    return gridi

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef inline int cutting_plane_cell(
                        np.float64_t x, np.float64_t y, np.float64_t z,
                        np.float64_t norm_vec[3], np.float64_t d,
                        np.float64_t dist):
    cdef np.float64_t cd = x*norm_vec[0] + y*norm_vec[1] + z*norm_vec[2] + d
    if fabs(cd) <= dist: return 1
    return 0

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def cutting_plane_cells(dobj, gobj):
    cdef np.ndarray[np.uint8_t, ndim=3] mask
    cdef np.ndarray[np.float64_t, ndim=1] left_edge = gobj.LeftEdge
    cdef np.ndarray[np.float64_t, ndim=1] dds = gobj.dds
    cdef int i, j, k
    cdef np.float64_t x, y, z, dist
    cdef np.float64_t norm_vec[3]
    cdef np.float64_t d = dobj._d

    mask = np.zeros(gobj.ActiveDimensions, dtype='uint8')
    for i in range(3): norm_vec[i] = dobj._norm_vec[i]
    dist = 0.5*(dds[0]*dds[0] + dds[1]*dds[1] + dds[2]*dds[2])**0.5
    x = left_edge[0] + dds[0] * 0.5
    for i in range(mask.shape[0]):
        y = left_edge[1] + dds[1] * 0.5
        for j in range(mask.shape[1]):
            z = left_edge[2] + dds[2] * 0.5
            for k in range(mask.shape[2]):
                mask[i,j,k] = cutting_plane_cell(x, y, z, norm_vec, d, dist)
                z += dds[1]
            y += dds[1]
        x += dds[0]
    return mask

# Finally, miscellaneous routines.

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_values_at_point(np.ndarray[np.float64_t, ndim=1] point,
                         np.ndarray[np.float64_t, ndim=2] left_edges,
                         np.ndarray[np.float64_t, ndim=2] right_edges,
                         np.ndarray[np.int32_t, ndim=2] dimensions,
                         field_names, grid_objects):
    # This iterates in order, first to last, and then returns with the first
    # one in which the point is located; this means if you order from highest
    # level to lowest, you will find the correct grid without consulting child
    # masking.  Note also that we will do a few relatively slow operations on
    # strings and whatnot, but they should not be terribly slow.
    cdef int ind[3], gi, fi
    cdef int nf = len(field_names)
    cdef np.float64_t dds
    cdef np.ndarray[np.float64_t, ndim=3] field
    cdef np.ndarray[np.float64_t, ndim=1] rv = np.zeros(nf, dtype='float64')
    for gi in range(left_edges.shape[0]):
        if not ((left_edges[gi,0] < point[0] < right_edges[gi,0])
            and (left_edges[gi,1] < point[1] < right_edges[gi,1])
            and (left_edges[gi,2] < point[2] < right_edges[gi,2])):
            continue
        # We found our grid!
        for fi in range(3):
            dds = ((right_edges[gi,fi] - left_edges[gi,fi])/
                   (<np.float64_t> dimensions[gi,fi]))
            ind[fi] = <int> ((point[fi] - left_edges[gi,fi])/dds)
        grid = grid_objects[gi]
        for fi in range(nf):
            field = grid[field_names[fi]]
            rv[fi] = field[ind[0], ind[1], ind[2]]
        return rv
    raise KeyError

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def obtain_rvec(data):
    # This is just to let the pointers exist and whatnot.  We can't cdef them
    # inside conditionals.
    cdef np.ndarray[np.float64_t, ndim=1] xf
    cdef np.ndarray[np.float64_t, ndim=1] yf
    cdef np.ndarray[np.float64_t, ndim=1] zf
    cdef np.ndarray[np.float64_t, ndim=2] rf
    cdef np.ndarray[np.float64_t, ndim=3] xg
    cdef np.ndarray[np.float64_t, ndim=3] yg
    cdef np.ndarray[np.float64_t, ndim=3] zg
    cdef np.ndarray[np.float64_t, ndim=4] rg
    cdef np.float64_t c[3]
    cdef int i, j, k
    center = data.get_field_parameter("center")
    c[0] = center[0]; c[1] = center[1]; c[2] = center[2]
    if len(data['x'].shape) == 1:
        # One dimensional data
        xf = data['x']
        yf = data['y']
        zf = data['z']
        rf = np.empty((3, xf.shape[0]), 'float64')
        for i in range(xf.shape[0]):
            rf[0, i] = xf[i] - c[0]
            rf[1, i] = yf[i] - c[1]
            rf[2, i] = zf[i] - c[2]
        return rf
    else:
        # Three dimensional data
        xg = data['x']
        yg = data['y']
        zg = data['z']
        rg = np.empty((3, xg.shape[0], xg.shape[1], xg.shape[2]), 'float64')
        for i in range(xg.shape[0]):
            for j in range(xg.shape[1]):
                for k in range(xg.shape[2]):
                    rg[0,i,j,k] = xg[i,j,k] - c[0]
                    rg[1,i,j,k] = yg[i,j,k] - c[1]
                    rg[2,i,j,k] = zg[i,j,k] - c[2]
        return rg

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def obtain_rv_vec(data):
    # This is just to let the pointers exist and whatnot.  We can't cdef them
    # inside conditionals.
    cdef np.ndarray[np.float64_t, ndim=1] vxf
    cdef np.ndarray[np.float64_t, ndim=1] vyf
    cdef np.ndarray[np.float64_t, ndim=1] vzf
    cdef np.ndarray[np.float64_t, ndim=2] rvf
    cdef np.ndarray[np.float64_t, ndim=3] vxg
    cdef np.ndarray[np.float64_t, ndim=3] vyg
    cdef np.ndarray[np.float64_t, ndim=3] vzg
    cdef np.ndarray[np.float64_t, ndim=4] rvg
    cdef np.float64_t bv[3]
    cdef int i, j, k
    bulk_velocity = data.get_field_parameter("bulk_velocity")
    if bulk_velocity == None:
        bulk_velocity = np.zeros(3)
    bv[0] = bulk_velocity[0]; bv[1] = bulk_velocity[1]; bv[2] = bulk_velocity[2]
    if len(data['x-velocity'].shape) == 1:
        # One dimensional data
        vxf = data['x-velocity'].astype("float64")
        vyf = data['y-velocity'].astype("float64")
        vzf = data['z-velocity'].astype("float64")
        rvf = np.empty((3, vxf.shape[0]), 'float64')
        for i in range(vxf.shape[0]):
            rvf[0, i] = vxf[i] - bv[0]
            rvf[1, i] = vyf[i] - bv[1]
            rvf[2, i] = vzf[i] - bv[2]
        return rvf
    else:
        # Three dimensional data
        vxg = data['x-velocity'].astype("float64")
        vyg = data['y-velocity'].astype("float64")
        vzg = data['z-velocity'].astype("float64")
        rvg = np.empty((3, vxg.shape[0], vxg.shape[1], vxg.shape[2]), 'float64')
        for i in range(vxg.shape[0]):
            for j in range(vxg.shape[1]):
                for k in range(vxg.shape[2]):
                    rvg[0,i,j,k] = vxg[i,j,k] - bv[0]
                    rvg[1,i,j,k] = vyg[i,j,k] - bv[1]
                    rvg[2,i,j,k] = vzg[i,j,k] - bv[2]
        return rvg



"""
Pluto-specific fields

Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2009-2011 J. S. Oishi, Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.data_objects.field_info_container import \
    FieldInfoContainer, \
    FieldInfo, \
    NullFunc, \
    ValidateParameter, \
    ValidateDataField, \
    ValidateProperty, \
    ValidateSpatial, \
    ValidateGridType
import yt.data_objects.universal_fields
import numpy as np

KnownPlutoFields = FieldInfoContainer()
add_pluto_field = KnownPlutoFields.add_field

PlutoFieldInfo = FieldInfoContainer.create_with_fallback(FieldInfo)
add_field = PlutoFieldInfo.add_field

add_pluto_field("rho", function=NullFunc, take_log=True,
                 validators = [ValidateDataField("density")],
                 units=r"\rm{g}/\rm{cm}^3")

KnownPlutoFields["rho"]._projected_units =r"\rm{g}/\rm{cm}^2"

add_pluto_field("vx1", function=NullFunc, take_log=False,
                 validators = [ValidateDataField("X-Momentum")],
                 units=r"",display_name=r"M_x")
KnownPlutoFields["vx1"]._projected_units=r""

add_pluto_field("vx2", function=NullFunc, take_log=False,
                 validators = [ValidateDataField("Y-Momentum")],
                 units=r"",display_name=r"M_y")
KnownPlutoFields["vx2"]._projected_units=r""

add_pluto_field("vx3", function=NullFunc, take_log=False,
                 validators = [ValidateDataField("Z-Momentum")],
                 units=r"",display_name=r"M_z")
KnownPlutoFields["vx3"]._projected_units=r""

add_pluto_field("prs", function=NullFunc, take_log=True,
                 validators = [ValidateDataField("energy-density")],
                 units=r"\rm{erg}/\rm{cm}^3")

def _Density(field,data):
    """A duplicate of the density field. This is needed because when you try 
    to instantiate a PlotCollection without passing in a center, the code
    will try to generate one for you using the "Density" field, which gives an error 
    if it isn't defined.

    """
    return data["rho"]
add_field("Density",function=_Density, take_log=True,
          units=r'\rm{g}/\rm{cm^3}')

def _Xmomentum(field, data):
    """ Generate x-momentum. """
    return data["vx1"]*data["density"]
add_field("X-momentum",function=_Xmomentum, take_log=False,
          units=r'\rm{g}/\rm{cm^2 s}')

def _Ymomentum(field, data):
    """ Generate y-momentum  """
    return data["vx2"]*data["density"]
add_field("Y-momentum",function=_Ymomentum, take_log=False,
          units=r'\rm{g}/\rm{cm^2 s}')

def _Zmomentum(field,data):
    """ Generate z-momentum"""
    return data["vx3"]*data["density"]
add_field("Z-Momentum",function=_Zmomentum, take_log=False,
          units=r'\rm{g}/\rm{cm^2 s}')




"""
The data-file handling functions

Author: Samuel W. Skillman <samskillman@gmail.com>
Affiliation: University of Colorado at Boulder
Author: Matthew Turk <matthewturk@gmail.com>
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from yt.utilities.io_handler import \
           BaseIOHandler
import numpy as np

class IOHandlerAthena(BaseIOHandler):
    _data_style = "athena"
    _offset_string = 'data:offsets=0'
    _data_string = 'data:datatype=0'
    _read_table_offset = None

    def _field_dict(self,fhandle):
        keys = fhandle['field_types'].keys()
        val = fhandle['field_types'].keys()
        return dict(zip(keys,val))

    def _read_field_names(self,grid):
        pass

    def _read_data(self,grid,field):
        f = file(grid.filename, 'rb')
        dtype, offsetr = grid.hierarchy._field_map[field]
        grid_ncells = np.prod(grid.ActiveDimensions)
        grid_dims = grid.ActiveDimensions
        grid0_ncells = np.prod(grid.hierarchy.grid_dimensions[0,:])
        read_table_offset = get_read_table_offset(f)
        if grid_ncells != grid0_ncells:
            offset = offsetr + ((grid_ncells-grid0_ncells) * (offsetr//grid0_ncells))
        if grid_ncells == grid0_ncells:
            offset = offsetr
        f.seek(read_table_offset+offset)
        if dtype == 'scalar':
            data = np.fromfile(f, dtype='>f4',
                    count=grid_ncells).reshape(grid_dims,order='F').copy()
        if dtype == 'vector':
            data = np.fromfile(f, dtype='>f4', count=3*grid_ncells)
            if '_x' in field:
                data = data[0::3].reshape(grid_dims,order='F').copy()
            elif '_y' in field:
                data = data[1::3].reshape(grid_dims,order='F').copy()
            elif '_z' in field:
                data = data[2::3].reshape(grid_dims,order='F').copy()
        f.close()
        if grid.pf.field_ordering == 1:
            return data.T
        else:
            return data

    def _read_data_slice(self, grid, field, axis, coord):
        sl = [slice(None), slice(None), slice(None)]
        sl[axis] = slice(coord, coord + 1)
        if grid.pf.field_ordering == 1:
            sl.reverse()
        return self._read_data_set(grid, field)[sl]


def get_read_table_offset(f):
    line = f.readline()
    while True:
        splitup = line.strip().split()
        if 'CELL_DATA' in splitup:
            f.readline()
            read_table_offset = f.tell()
            break
        line = f.readline()
    return read_table_offset





"""
Calculate the radial column density around a point.

Author: Stephen Skory <s@skory.us>
Affiliation: CU Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Stephen Skory.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from yt.mods import *
import yt.visualization.volume_rendering.camera as camera
import yt.utilities.lib as au
from yt.utilities.math_utils import periodic_dist
from yt.data_objects.field_info_container import FieldDetector
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    ParallelAnalysisInterface

def _col_dens(field, data):
    return data[field]

class RadialColumnDensity(ParallelAnalysisInterface):
    r"""
    Calculate radial column densities in preparation to
    adding them as a derived field.
    
    This class is the first step in calculating a derived radial 
    column density field.
    Given a central point, this calculates the column density to all cell
    centers within the given radius in units of centimeters times
    the units of the basis field.
    For example, if the basis field is `NumberDensity`, which has units
    of 1 / cm^3, the units of the derived field will be 1 / cm^2.
    Please see the documentation or the example below on how to
    use this to make the derived field which can be used like any other
    derived field.
    
    This builds a number of spherical 
    surfaces where the column density is calculated
    using HEALPix Volume Rendering. The values of the column density at
    grid points is then linearly interpolated between the two nearest
    surfaces (one inward, one outward).
    Please see the HEALPix Volume Rendering documentation for more on
    that part of this calculation.
    
    Parameters
    ----------
    pf : `StaticOutput`
        The dataset to operate on.
    field : string
        The name of the basis field over which to
        calculate a column density.
    center : array_like
        A list or array giving the location of where to
        calculate the start of
        the column density.
        This will probably be "an object of interest" like
        a star, black hole, or the center of a galaxy.
    max_radius : float
        How far out to calculate the column density, in code units. This
        value will be automatically reduced if the supplied value would
        result in calculating column densities outside the volume.
        Default = 0.5.
    steps : integer
        How many surfaces to use. A higher number is more accurate, but
        takes more resources.
        Default = 10
    base : string
        How to evenly space the surfaces: linearly with "lin" or
        logarithmically with "log".
        Default = "lin".
    Nside : int
        The resolution of column density calculation as performed by
        HEALPix. Higher numbers mean higher quality. Max = 8192.
        Default = 32.
    ang_divs : imaginary integer
        This number controls the gridding of the HEALPix projection onto
        the spherical surfaces. Higher numbers mean higher quality.
        Default = 800j.
    
    Examples
    --------
    
    >>> rcdnumdens = RadialColumnDensity(pf, 'NumberDensity', [0.5, 0.5, 0.5])
    >>> def _RCDNumberDensity(field, data, rcd = rcdnumdens):
            return rcd._build_derived_field(data)
    >>> add_field('RCDNumberDensity', _RCDNumberDensity, units=r'1/\rm{cm}^2')
    """
    def __init__(self, pf, field, center, max_radius = 0.5, steps = 10,
            base='lin', Nside = 32, ang_divs = 800j):
        ParallelAnalysisInterface.__init__(self)
        self.pf = pf
        self.center = np.asarray(center)
        self.max_radius = max_radius
        self.steps = steps
        self.base = base
        self.Nside = Nside
        self.ang_divs = ang_divs
        self.real_ang_divs = int(np.abs(ang_divs))
        self.phi, self.theta = np.mgrid[0.0:2*np.pi:ang_divs, 0:np.pi:ang_divs]
        self.phi1d = self.phi[:,0]
        self.theta1d = self.theta[0,:]
        self.dphi = self.phi1d[1] - self.phi1d[0]
        self.dtheta = self.theta1d[1] - self.theta1d[0]
        self.pixi = au.arr_ang2pix_nest(Nside, self.theta.ravel(), self.
            phi.ravel())
        self.dw = pf.domain_right_edge - pf.domain_left_edge
        # Here's where we actually do stuff.
        self._fix_max_radius()
        self._make_bins()
        self._build_surfaces(field)
    
    def _fix_max_radius(self):
        # The max_radius can only be the distance from the center point to
        # the closest face of the volume. This is because the column density
        # for a surface outside the volume is ill-defined due to the way 
        # normalization is handled in the volume render.
        # It may be possible to fix this in
        # the future, and allow these calculations in the whole volume,
        # but this will work for now.
        right = self.pf.domain_right_edge - self.center
        left = self.center - self.pf.domain_left_edge
        min_r = np.min(right)
        min_l = np.min(left)
        self.max_radius = np.min([self.max_radius, min_r, min_l])
    
    def _make_bins(self):
        # We'll make the bins start from the smallest cell size to the
        # specified radius. Column density inside the same cell as our 
        # center is kind of ill-defined, anyway.
        if self.base == 'lin':
            self.bins = np.linspace(self.pf.h.get_smallest_dx(), self.max_radius,
                self.steps)
        elif self.base == 'log':
            self.bins = np.logspace(np.log10(self.pf.h.get_smallest_dx()),
                np.log10(self.max_radius), self.steps)
    
    def _build_surfaces(self, field):
        # This will be index by bin index.
        self.surfaces = {}
        for i, radius in enumerate(self.bins):
            cam = camera.HEALpixCamera(self.center, radius, self.Nside,
                pf = self.pf, log_fields = [False], fields = field)
            bitmap = cam.snapshot()
            self.surfaces[i] = radius * self.pf['cm'] * \
                bitmap[:,0,0][self.pixi].reshape((self.real_ang_divs,self.real_ang_divs))
            self.surfaces[i] = self.comm.mpi_allreduce(self.surfaces[i], op='max')

    def _build_derived_field(self, data, minval=None):
        r"""
        Parameters
        ----------
        
        minval : float
            This parameter will set any values of the
            field that are zero to this minimum value.
            Values of zero are found outside the maximum radius and
            in the cell of the user-specified center point.
            This setting is useful if the field is going to be logged
            (e.g. np.log10) where zeros are inconvenient.
            Default = None
        """
        x = data['x']
        sh = x.shape
        ad = np.prod(sh)
        if type(data) == type(FieldDetector()):
            return np.ones(sh)
        y = data['y']
        z = data['z']
        pos = np.array([x.reshape(ad), y.reshape(ad), z.reshape(ad)]).T
        del x, y, z
        vals = self._interpolate_value(pos)
        del pos
        if minval:
            zeros = (vals == 0.)
            vals[zeros] = minval
            del zeros
        vals.shape = sh
        return vals
    
    def _interpolate_value(self, pos):
        # Given a position, find the two surfaces it's in between,
        # and the interpolate values from the surfaces to the point
        # according to the points angle.
        # 1. Find the angle from the center point to the position.
        vec = pos - self.center
        phi = np.arctan2(vec[:, 1], vec[:, 0])
        # Convert the convention from [-pi, pi) to [0, 2pi).
        sel = (phi < 0)
        phi[sel] += 2 * np.pi
        # Find the radius.
        r = np.sqrt(np.sum(vec * vec, axis = 1))
        # Keep track of the points outside of self.max_radius, which we'll
        # handle separately before we return.
        outside = (r > self.max_radius)
        theta = np.arccos(vec[:, 2] / r)
        # 2. Find the bin for this position.
        digi = np.digitize(r, self.bins)
        # Find the values on the inner and outer surfaces.
        in_val = np.zeros_like(r)
        out_val = np.zeros_like(r)
        # These two will be used for interpolation.
        in_r = np.zeros_like(r)
        out_r = np.zeros_like(r)
        for bin in np.unique(digi):
            sel = (digi == bin)
            # Special case if we're outside the largest sphere.
            if bin == len(self.bins):
                in_val[sel] = 0.
                out_val[sel] = 0.
                # Just something non-zero so we don't get divide errors later.
                in_r[sel] = .1
                out_r[sel] = .2
                continue
            # Special case if we're inside the smallest sphere.
            elif bin == 0:
                in_val[sel] = np.zeros_like(phi[sel])
                in_r[sel] = 0.
                out_val[sel] = self._interpolate_surface_value(1,
                    phi[sel], theta[sel])
                out_r[sel] = self.bins[1]
                continue
            # General case.
            else:
                in_val[sel] = self._interpolate_surface_value(bin - 1,
                    phi[sel], theta[sel])
                in_r[sel] = self.bins[bin - 1]
                out_val[sel] = self._interpolate_surface_value(bin,
                    phi[sel], theta[sel])
                out_r[sel] = self.bins[bin]
        # Interpolate using a linear fit in column density / r space.
        val = np.empty_like(r)
        # Special case for inside smallest sphere.
        sel = (digi == 0)
        val[sel] = (1. - (out_r[sel] - r[sel]) / out_r[sel]) * out_val[sel]
        np.invert(sel, sel) # In-place operation!
        val[sel] = (out_val[sel] - in_val[sel]) / (out_r[sel] - in_r[sel]) * \
            (r[sel] - in_r[sel]) + in_val[sel]
        # Fix the things to zero that should be zero.
        val[outside] = 0.
        return val
        
    def _interpolate_surface_value(self, bin, phi, theta):
        # Given a surface bin and an angle, interpolate the value on
        # that surface to the angle.
        # 1. Find the four values closest to the angle.
        phi_bin = np.digitize(phi, self.phi1d)
        theta_bin = np.digitize(theta, self.theta1d)
        val00 = self.surfaces[bin][phi_bin - 1, theta_bin - 1]
        val01 = self.surfaces[bin][phi_bin - 1, theta_bin]
        val10 = self.surfaces[bin][phi_bin, theta_bin - 1]
        val11 = self.surfaces[bin][phi_bin, theta_bin]
        # 2. Linearly interpolate the four values to the points.
        int_val0 = (val10 - val00) / self.dphi * \
            (phi - self.phi1d[phi_bin - 1]) + val00
        int_val1 = (val11 - val01) / self.dphi * \
            (phi - self.phi1d[phi_bin - 1]) + val01
        vals = (int_val1 - int_val0) / self.dtheta * \
            (theta - self.theta1d[theta_bin - 1]) + int_val0
        return vals


